from pathlib								import Path
from typing									import List
from typing									import Tuple
from typing									import Optional
from typing									import Generator
from pygwarts.magical.philosophers_stone	import Transmutable
from pygwarts.hagrid.thrivables				import Tree
from pygwarts.hagrid.thrivables				import Copse
from pygwarts.hagrid.bloom					import Bloom
from pygwarts.hagrid.cultivation.sifting	import SiftingController








def unplantable(*plant :Tuple[str, Path, List[Path], List[Path]]) -> None | str :

	"""
		Helper function that validates "plant" consistency.
		All collected "plant" items must represent a tuple of four objects:
			1. a string representing "sprout" for current plant;
			2. a Path object representing current branch for current plant;
			3. list of Path objects representing current branch twigs for current plant;
			4. list of Path objects representing current branch leafs for current plant.
		If "plant" fully suffices returns None, otherwise the string message what is wrong.
	"""

	if	(pcount := len(plant)) != 4:		return	f"Invalid plant length {pcount}"
	elif(not isinstance(plant[0], str)):	return	f"Invalid plant sprout \"{plant[0]}\""
	elif(not isinstance(plant[1], Path)):	return	f"Invalid plant branch \"{plant[1]}\""
	elif(not isinstance(plant[2], list)):	return	f"Invalid plant twigs {type(plant[2])}"
	elif(not isinstance(plant[3], list)):	return	f"Invalid plant leafs {type(plant[3])}"








class Flourish(Transmutable):

	"""
		hagrid core functionality class, that serves as a "planting dispatcher", which means the Flourish
		object is responsible for every Tree object processing.
		Flourish implementation suggests __call__ will accept the Tree object that represent current target
		object and a tuple of four objects: string object representing current sprout (source root folder),
		Path object representing branch (current source folder to be processed) and two lists of Path objects,
		that are twigs and leafs (current source folder folders and files).
		This is mandatory interface and Folurish is guarded by precheks. In __call__ "unplantable" helper
		function is used to check "plant" consistency, so every "plant", that not a tuple of valid four, will
		be skipped. Such strict interface requirements must ensure right usage, as was originally designed.
		For valid "plant" there are three steps of "planting" (processing) in Flourish:
			1.	Global sifting phase (or first step filter). This is the main folders/files filtering, that
		will affect every Tree object participated in Flourish. For this phase there must be SiftingController
		objects to be reachable for Flourish object (declared in escalation) - "twigs" for folders and
		"leafs" for files. The best way to declare them is right under the Flourish object, cause that will
		ensure reachability and also isolation from any Tree/Copse objects. Any or both might be omitted.
			2.	Local sifting phase (or Tree objects extra sifting). After global sifting, "innervate" will
		start it's primary job - starting from the "origin" argument, that must be the upper layer Copse or
		Tree object, "innervate" will find every Tree object in vicinity, iterating over upper layer Copse
		object and it's Copse objects respectively, in order to start "planting" it (processing). This phase
		is transitional, means it allows the start of final third phase for every Tree object found, but
		before this will happen, "innervate" invokes SiftingController objects for every Tree object as extra
		filtering. As every valid Tree objects implemented to have 2 special methods "twigify" for "twigs"
		and "leafify" for leafs, those methods will try to invoke "controllers" if they will find some.
			3. Trees blooming (processing every reachable Tree object). Every Tree object, that yielded by
		"grove" helper method, will be iterated and every Bloom object assigned to its Tree will be invoked.
		This invocation relies every Bloom object implements "bloom-interface":
			accepts:
				origin	- current Tree object to be processed;
				sprout	- current source root folder as a string;
				branch	- current source working directory as a Path object;
				twigs	- list of Path objects as current source folders;
				leafs	- list of Path objects as current source files.
			return:
				None.
		For Flourish object to get it's job done, it's upper layer must be a single Tree or an upper Copse
		object, that encapsulates all possible Tree and other Copse objects.
	"""


	twigs	:Optional[SiftingController]
	leafs	:Optional[SiftingController]


	def grove(self, origin :Copse | Tree) -> Generator[Tree, None,None] | None:

		"""
			Helper method that searches for Tree objects in some scope.
			The only argument "origin" serves as a entering point for search when it's first time called.
			If it is a Tree object, it is yielded and "grove" stops. If it is a Copse object, it will be
			iterated and every Tree object found will be yielded, while every nested Copse object will be
			used as "origin" for recursive call. Any other object is skipped and logged with warning.
		"""

		match origin:

			case Copse():
				for inner in origin:

					self.loggy.debug(f"Walking copse {origin} inner {inner}")
					yield	from	self.grove(inner)
			case Tree():	yield	origin
			case _:			self.loggy.warning(f"Invalid thrivable \"{origin}\" for flourish")








	def innervate(self, origin :Copse | Tree, *plant :Tuple[str, Path, List[Path], List[Path]]):

		"""
			Flourish core method that responsible for innervating "planting" for every Tree object.
			Accepts first positional argument "origin", which must refer to the entering point for "grove"
			helper method. All further arguments will be collected as "plant" and considered to be a tuple
			of sprout (source root folder) as a string, current branch as Path object, current branch
			twigs and leafs as lists of Path objects. The "plant" validation must occur beforehand, as it
			is in __call__ by the use of "unplantable" helper function.
			Consists of three phases:
				1.	Global sifting phase (or first step filter). This is the main folders/files filtering,
			that will affect every Tree object. For this phase there must be SiftingController objects
			to be reachable.
				2.	Local sifting phase (or Tree objects extra sifting). After global sifting, "innervate"
			will start it's primary job - starting from the "origin" argument, that must be the upper layer
			Copse or Tree object, "innervate" will find every Tree object in vicinity, iterating over upper
			layer Copse object and it's Copse objects respectively, in order to start "planting" it. This
			phase is transitional, means it allows the start of final third phase for every Tree object
			found, but before this will happen, "innervate" invokes SiftingController objects for every Tree
			object as extra filtering.
				3. Trees blooming (processing every reachable Tree object). Every Tree object, that yielded
			by "grove" helper method, will be iterated and every Bloom object assigned to its Tree will be
			invoked. This invocation relies on every Bloom object implements "bloom-interface":
				accepts:
					origin	- current Tree object to be processed;
					sprout	- current source root folder as a string;
					branch	- current source working directory as a Path object;
					twigs	- list of Path objects as current source folders;
					leafs	- list of Path objects as current source files.
				return:
					None.
		"""


		sprout, branch, branch_twigs, branch_leafs = plant
		branch_twigs_count = len(branch_twigs)
		branch_leafs_count = len(branch_leafs)


		self.loggy.debug(f"Current branch \"{branch}\"")
		self.loggy.debug(f"Number of twigs: {branch_twigs_count}")
		self.loggy.debug(f"Number of leafs: {branch_leafs_count}")




		# Twigs main sifting, or folders first step filter. This is the only place where
		# SiftingController invoked with two boolean flags toggled to True.
		# "thriving" turned on means after a branch failed to "fullmatch" the only includable,
		# there will be a check if the whole branch is a prefix of that includable. As this
		# technic is used only for twigs sifting, it allows to not sift out the branch that
		# would lead to included targets. The necessity to keep that twigs is related to
		# second toggled flag "popping", which means that sifting twigs happens literally.
		# When the twig is sifted out it is popped from the input list, so "plant" Generator
		# will refer to modified list of folders, where there'll be no folders, that are
		# either not included or excluded, so the walk will not occur, hence no Flourish
		# for that branches.
		if	callable((twigs_sift := getattr(self, "twigs", None))):
			current_twigs = twigs_sift(branch_twigs, thriving=True, popping=True)

			if	(twigified := len(current_twigs)) != branch_twigs_count:

				self.loggy.debug(f"Twigs after sifting: {twigified}")
		else:	current_twigs = branch_twigs


		# Leafs of current branch sifting that is general for every bough, so it must be clear
		# that include and exclude fields for this controller will affect every Tree Flourish.
		if	callable((leafs_sift := getattr(self, "leafs", None))):
			current_leafs = leafs_sift(branch_leafs)

			if	(leafified := len(current_leafs)) != branch_leafs_count:

				self.loggy.debug(f"Leafs after sifting: {leafified}")
		else:	current_leafs = branch_leafs




		for tree in self.grove(origin):
			if	isinstance(tree, Tree):


				# Twigs extra sifting of every Tree object.
				# This must be pointed, that twigs sifting for every Tree object, that is
				# to be escalated "SiftingController", is invoked with "thriving" flag on
				# and "popping" flag off. The first one means that it is, like in "Flourish"
				# twigs sifting, every walked branch will be checked to be a prefix for any
				# includable, which is the way to include final branch in pattern. The second
				# one means the list of twigs will not be altered (no pops) by sifting twigs.
				# Unlike the global "Flourish" twigs sifting, where "popping" is set on by
				# default to ensure sifting twigs from walking, local Tree sifting is actually
				# filter for twigs thriving (Germination). For purposes of filtering files
				# from certain folders, "leafs" sifting must be used.
				tree_twigs = tree.twigify(current_twigs)
				tree_leafs = tree.leafify(current_leafs)


				for bloom in tree:
					if	isinstance(bloom, Bloom):


						self.loggy.debug(f"Invoking {tree} tree {bloom} blooming")
						bloom(tree, sprout, branch, tree_twigs, tree_leafs)
					else:
						self.loggy.debug(f"Invalid Bloom \"{bloom}\" to flourish")




	def __call__(self, *plant :Tuple[str, Path, List[Path], List[Path]]):


		if	hasattr(self, "_UPPER_LAYER"):
			if	plant:

				if	isinstance(reason := unplantable(*plant), str): self.loggy.warning(reason)
				else:

					self.loggy.debug(f"Flourishing {plant[0]}")
					self.innervate(self._UPPER_LAYER, *plant)
			else:	self.loggy.debug(f"No plants to flourish")
		else:		self.loggy.critical(f"{self} failed to start flourish")










from setuptools import setup
from setuptools import find_packages








with open("README.md") as md : long_desc = md.read()
setup(

	name="pygwarts",
	version="1.4.9.6",
	author="lngd",
	author_email="lngdeer@gmail.com",
	url="https://github.com/longdeer/pygwarts",
	license="GPL-3.0 license",
	description="Python multitool library",
	long_description=long_desc,
	long_description_content_tpye="text/markdown",
	packages=find_packages(),
	python_requires=">=3.10"
)


import	os
from	pathlib			import Path
from	pygwarts.tests	import PygwartsTestCase








class MagicalTestCase(PygwartsTestCase):

	"""
		pygwarts.magical tests super class
	"""


	MAGICAL_ROOT			= Path(PygwartsTestCase._PygwartsTestCase__CWD) /"magical"
	TIMETURN_TIMERS_HANDLER	= str(Path(PygwartsTestCase._PygwartsTestCase__CWD) /"magical" /"timers.loggy")
	KEY_SET_HANDLER			= str(Path(PygwartsTestCase._PygwartsTestCase__CWD) /"magical" /"kchest.loggy")
	PACKED_KEY_SET_HANDLER	= str(Path(PygwartsTestCase._PygwartsTestCase__CWD) /"magical" /"pkchest.loggy")
	DEEP_KEY_SET_HANDLER	= str(Path(PygwartsTestCase._PygwartsTestCase__CWD) /"magical" /"dkchest.loggy")
	SET_HANDLER				= str(Path(PygwartsTestCase._PygwartsTestCase__CWD) /"magical" /"chest.loggy")
	PACKED_SET_HANDLER		= str(Path(PygwartsTestCase._PygwartsTestCase__CWD) /"magical" /"pchest.loggy")


	HOURS	= tuple(range(24))
	MINSEC	= tuple(range(60))
	DAYS_28	= tuple(range(1,29))
	DAYS_29	= tuple(range(1,30))
	DAYS_30	= tuple(range(1,31))
	DAYS_31	= tuple(range(1,32))
	MONTHS	= tuple(range(1,13))
	YEARS	= tuple(range(16,26))
	FYEARS	= tuple(range(2020,2026))

	DELIMETERS = ",", "\\", "/", ".", "-", ":", " ",

	travels = {

		"microseconds":	100500,
		"milliseconds":	100500,
		"seconds":		420,
		"minutes":		69,
		"hours":		42,
		"days":			17,
		"weeks":		14,
	}

	months = (

		"january",
		"february",
		"march",
		"april",
		"may",
		"june",
		"july",
		"august",
		"september",
		"october",
		"november",
		"december"
	)










import	re
from	pygwarts.magical.philosophers_stone	import Transmutable








EMAIL_REGEX_PATTERN = re.compile(

	r"([-!#-'*+/-9=?A-Z^-~]+(\.[-!#-'*+/-9=?A-Z^-~]+)*|\"([]!#-[^-~ \t]|(\\[\t -~]))+\")"
	"@"
	r"[0-9A-Za-z]([0-9A-Za-z-]{0,61}[0-9A-Za-z])?(\.[0-9A-Za-z]([0-9A-Za-z-]{0,61}[0-9A-Za-z])?)+"
)








class EmailValidator(Transmutable):

	"""
		Utility class to validate email address. Accepts string, that must represent an email to
		validate. Returns a string argument "address" if it is valid email address, None otherwise - if
		validation failed or if "email" argument is not a string.
		Validation goes by "EMAIL_REGEX_PATTERN" regex, that is:
			1.	have a local part (part before the @-sign) that is strictly compliant with RFC 5321/5322;
			2.	have a domain part (part after the @-sign) that is a host name with at least two labels,
				each of which is at most 63 characters long.
	"""

	def __call__(self, email :str) -> str | None :

		if	isinstance(email, str):
			if	EMAIL_REGEX_PATTERN.fullmatch(email):

				self.loggy.debug(f"Email address \"{email}\" validated")
				return	email

		self.loggy.info(f"Invalid email address \"{email}\"")










import	os
import	unittest
from	shutil								import rmtree
from	time								import sleep
from	pygwarts.tests.hagrid				import MediumSet
from	pygwarts.irma.contrib				import LibraryContrib
from	pygwarts.hagrid.thrivables			import Tree
from	pygwarts.hagrid.sprouts				import fssprout
from	pygwarts.hagrid.planting			import Flourish
from	pygwarts.hagrid.bloom.twigs			import Germination
from	pygwarts.hagrid.bloom.leafs			import Rejuvenation
from	pygwarts.hagrid.planting.leafs		import LeafGrowth
from	pygwarts.hagrid.planting.twigs		import TwigThrive
from	pygwarts.hagrid.planting.peels		import GrowingPeel
from	pygwarts.hagrid.planting.peeks		import DraftPeek
from	pygwarts.hagrid.cultivation.sifting	import SiftingController








class MultisproutTree(MediumSet):

	"""
		GrowingPeel, DraftPeek, Germination -> Rejuvenation
		Triple sprout copse
		Single bough tree
	"""

	@classmethod
	def tearDownClass(cls):
		cls.clean(cls)

		if	cls.clean_up:
			if	os.path.isfile(cls.MEDIUM_HANDLER_4): os.remove(cls.MEDIUM_HANDLER_4)

	@classmethod
	def setUpClass(cls):

		super().setUpClass()
		cls.make_loggy_file(cls, cls.MEDIUM_HANDLER_4)
		if	os.path.isdir(cls.MEDIUM_SET_BOUGH_2): rmtree(cls.MEDIUM_SET_BOUGH_2)
		if	os.path.isdir(cls.MEDIUM_SET_BOUGH_3): rmtree(cls.MEDIUM_SET_BOUGH_3)

	def setUp(self):
		class Sakura(Tree):

			bough = self.MEDIUM_SET_BOUGH_1
			class loggy(LibraryContrib):

				handler		= self.MEDIUM_HANDLER_4
				init_level	= 10


			@GrowingPeel
			class thrive(TwigThrive):		pass
			class folders(Germination):		pass

			@GrowingPeel
			@DraftPeek(renew=False, picky=True)
			class grow(LeafGrowth):			pass
			class files(Rejuvenation):		pass

			@fssprout(self.MEDIUM_SET_SPROUT_3)
			@fssprout(self.MEDIUM_SET_SPROUT_2)
			@fssprout(self.MEDIUM_SET_SPROUT_1)
			class sync(Flourish):

				class twigs(SiftingController):

					# no bads, please (missing not bad)
					exclude	= ( rf".+/.*bad.*$", ) if os.name == "posix" else ( rf".+\\.*bad.*$", )

				class leafs(SiftingController):

					# hate crumbs?
					exclude	= ( rf".+/crumb.*", ) if os.name == "posix" else ( rf".+\\crumb.*", )
					# include BOUGH_1 files for trimming
					# include SPROUT_1 files for sync
					# include SPROUT_3 files for sync
					include	= (
						(
							rf"{self.MEDIUM_SET_BOUGH_1}{os.sep}.+",
							rf"{self.MEDIUM_SET_SPROUT_1}{os.sep}.+",
							rf"{self.MEDIUM_SET_SPROUT_3}{os.sep}.+",
						)	if os.name == "posix" else
						(
							self.MEDIUM_SET_BOUGH_1.replace("\\", "\\\\") + r"\\.+",
							self.MEDIUM_SET_SPROUT_1.replace("\\", "\\\\") + r"\\.+",
							self.MEDIUM_SET_SPROUT_3.replace("\\", "\\\\") + r"\\.+",
						)
					)


		self.test_case = Sakura




	def test_first_flourish(self):

		self.test_case.loggy.init_name = "first_flourish"
		self.assertTrue(os.path.isdir(self.MEDIUM_SET_BOUGH_1))


		# First sprout
		self.assertTrue(os.path.isfile(self.f_mss1_md_rock))
		self.assertTrue(os.path.isfile(self.f_mss1_md_paper))
		self.assertTrue(os.path.isfile(self.f_mss1_md_scissors))
		self.assertTrue(os.path.isfile(self.f_mss1_md_gc_sadmovies))
		self.assertTrue(os.path.isfile(self.f_mss1_md_gc_dumbasses))
		self.assertTrue(os.path.isfile(self.f_mss1_md_gc_onion))
		self.assertTrue(os.path.isdir(self.d_mss1_c_good))
		self.assertTrue(os.path.isdir(self.d_mss1_c_notgood))
		self.assertTrue(os.path.isdir(self.d_mss1_c_bad))
		self.assertTrue(os.path.isdir(self.d_mss1_c_notbad))
		self.assertTrue(os.path.isdir(self.d_mss1_c_notnot))
		self.assertTrue(os.path.isdir(self.d_mss1_c_badbad))
		self.assertTrue(os.path.isdir(self.d_mss1_c_badgood))
		self.assertTrue(os.path.isdir(self.d_mss1_c_goodbad))

		# Second sprout
		self.assertTrue(os.path.isfile(self.f_mss2_tasksschedule))
		self.assertTrue(os.path.isfile(self.f_mss2_explangraph))
		self.assertTrue(os.path.isfile(self.f_mss2_c_hardcleaning))
		self.assertTrue(os.path.isfile(self.f_mss2_c_hardwashing))
		self.assertTrue(os.path.isfile(self.f_mss2_c_mediumcooking))
		self.assertTrue(os.path.isfile(self.f_mss2_c_mediumhomework))
		self.assertTrue(os.path.isfile(self.f_mss2_c_mediumphysicals))
		self.assertTrue(os.path.isfile(self.f_mss2_c_easyprocrastinating))
		self.assertTrue(os.path.isfile(self.f_mss2_c_easyreflections))
		self.assertTrue(os.path.isfile(self.f_mss2_c_easyphilosophizing))

		# Third sprout
		self.assertTrue(os.path.isfile(self.f_mss3_k_f_milk))
		self.assertTrue(os.path.isfile(self.f_mss3_k_f_cheese))
		self.assertTrue(os.path.isfile(self.f_mss3_k_f_meat))
		self.assertTrue(os.path.isfile(self.f_mss3_k_o_chicken))
		self.assertTrue(os.path.isfile(self.f_mss3_k_o_pie))
		self.assertTrue(os.path.isfile(self.f_mss3_k_t_bc_bread))
		self.assertTrue(os.path.isfile(self.f_mss3_k_t_bc_crumb1))
		self.assertTrue(os.path.isfile(self.f_mss3_k_t_bc_crumb2))
		self.assertTrue(os.path.isfile(self.f_mss3_k_t_bc_crumb420))
		self.assertTrue(os.path.isfile(self.f_mss3_k_t_bc_crumb69))




		with self.assertLogs("first_flourish", 10) as case_loggy:

			self.test_case = self.test_case()
			self.test_case.sync()




		self.no_loggy_levels(case_loggy.output, 30,40,50)


		self.assertIn(

			f"DEBUG:first_flourish:New draft peek \"{self.f_mss1_md_rock}\" "
			f"new ring: {os.path.getmtime(self.f_mss1_md_rock)}",
			case_loggy.output
		)
		self.assertIn(

			f"DEBUG:first_flourish:New draft peek \"{self.f_mss1_md_paper}\" "
			f"new ring: {os.path.getmtime(self.f_mss1_md_paper)}",
			case_loggy.output
		)
		self.assertIn(

			f"DEBUG:first_flourish:New draft peek \"{self.f_mss1_md_scissors}\" "
			f"new ring: {os.path.getmtime(self.f_mss1_md_scissors)}",
			case_loggy.output
		)
		self.assertIn(

			f"DEBUG:first_flourish:New draft peek \"{self.f_mss1_md_gc_sadmovies}\" "
			f"new ring: {os.path.getmtime(self.f_mss1_md_gc_sadmovies)}",
			case_loggy.output
		)
		self.assertIn(

			f"DEBUG:first_flourish:New draft peek \"{self.f_mss1_md_gc_dumbasses}\" "
			f"new ring: {os.path.getmtime(self.f_mss1_md_gc_dumbasses)}",
			case_loggy.output
		)
		self.assertIn(

			f"DEBUG:first_flourish:New draft peek \"{self.f_mss1_md_gc_onion}\" "
			f"new ring: {os.path.getmtime(self.f_mss1_md_gc_onion)}",
			case_loggy.output
		)




		self.assertNotIn(

			f"DEBUG:first_flourish:New draft peek \"{self.f_mss2_tasksschedule}\" "
			f"new ring: {os.path.getmtime(self.f_mss2_tasksschedule)}",
			case_loggy.output
		)
		self.assertNotIn(

			f"DEBUG:first_flourish:New draft peek \"{self.f_mss2_explangraph}\" "
			f"new ring: {os.path.getmtime(self.f_mss2_explangraph)}",
			case_loggy.output
		)
		self.assertNotIn(

			f"DEBUG:first_flourish:New draft peek \"{self.f_mss2_c_hardcleaning}\" "
			f"new ring: {os.path.getmtime(self.f_mss2_c_hardcleaning)}",
			case_loggy.output
		)
		self.assertNotIn(

			f"DEBUG:first_flourish:New draft peek \"{self.f_mss2_c_hardwashing}\" "
			f"new ring: {os.path.getmtime(self.f_mss2_c_hardwashing)}",
			case_loggy.output
		)
		self.assertNotIn(

			f"DEBUG:first_flourish:New draft peek \"{self.f_mss2_c_mediumcooking}\" "
			f"new ring: {os.path.getmtime(self.f_mss2_c_mediumcooking)}",
			case_loggy.output
		)
		self.assertNotIn(

			f"DEBUG:first_flourish:New draft peek \"{self.f_mss2_c_mediumhomework}\" "
			f"new ring: {os.path.getmtime(self.f_mss2_c_mediumhomework)}",
			case_loggy.output
		)
		self.assertNotIn(

			f"DEBUG:first_flourish:New draft peek \"{self.f_mss2_c_mediumphysicals}\" "
			f"new ring: {os.path.getmtime(self.f_mss2_c_mediumphysicals)}",
			case_loggy.output
		)
		self.assertNotIn(

			f"DEBUG:first_flourish:New draft peek \"{self.f_mss2_c_easyprocrastinating}\" "
			f"new ring: {os.path.getmtime(self.f_mss2_c_easyprocrastinating)}",
			case_loggy.output
		)
		self.assertNotIn(

			f"DEBUG:first_flourish:New draft peek \"{self.f_mss2_c_easyreflections}\" "
			f"new ring: {os.path.getmtime(self.f_mss2_c_easyreflections)}",
			case_loggy.output
		)
		self.assertNotIn(

			f"DEBUG:first_flourish:New draft peek \"{self.f_mss2_c_easyphilosophizing}\" "
			f"new ring: {os.path.getmtime(self.f_mss2_c_easyphilosophizing)}",
			case_loggy.output
		)




		self.assertIn(

			f"DEBUG:first_flourish:New draft peek \"{self.f_mss3_k_f_milk}\" "
			f"new ring: {os.path.getmtime(self.f_mss3_k_f_milk)}",
			case_loggy.output
		)
		self.assertIn(

			f"DEBUG:first_flourish:New draft peek \"{self.f_mss3_k_f_cheese}\" "
			f"new ring: {os.path.getmtime(self.f_mss3_k_f_cheese)}",
			case_loggy.output
		)
		self.assertIn(

			f"DEBUG:first_flourish:New draft peek \"{self.f_mss3_k_f_meat}\" "
			f"new ring: {os.path.getmtime(self.f_mss3_k_f_meat)}",
			case_loggy.output
		)
		self.assertIn(

			f"DEBUG:first_flourish:New draft peek \"{self.f_mss3_k_o_chicken}\" "
			f"new ring: {os.path.getmtime(self.f_mss3_k_o_chicken)}",
			case_loggy.output
		)
		self.assertIn(

			f"DEBUG:first_flourish:New draft peek \"{self.f_mss3_k_o_pie}\" "
			f"new ring: {os.path.getmtime(self.f_mss3_k_o_pie)}",
			case_loggy.output
		)
		self.assertIn(

			f"DEBUG:first_flourish:New draft peek \"{self.f_mss3_k_t_bc_bread}\" "
			f"new ring: {os.path.getmtime(self.f_mss3_k_t_bc_bread)}",
			case_loggy.output
		)
		self.assertNotIn(

			f"DEBUG:first_flourish:New draft peek \"{self.f_mss3_k_t_bc_crumb1}\" "
			f"new ring: {os.path.getmtime(self.f_mss3_k_t_bc_crumb1)}",
			case_loggy.output
		)
		self.assertNotIn(

			f"DEBUG:first_flourish:New draft peek \"{self.f_mss3_k_t_bc_crumb2}\" "
			f"new ring: {os.path.getmtime(self.f_mss3_k_t_bc_crumb2)}",
			case_loggy.output
		)
		self.assertNotIn(

			f"DEBUG:first_flourish:New draft peek \"{self.f_mss3_k_t_bc_crumb420}\" "
			f"new ring: {os.path.getmtime(self.f_mss3_k_t_bc_crumb420)}",
			case_loggy.output
		)
		self.assertNotIn(

			f"DEBUG:first_flourish:New draft peek \"{self.f_mss3_k_t_bc_crumb69}\" "
			f"new ring: {os.path.getmtime(self.f_mss3_k_t_bc_crumb69)}",
			case_loggy.output
		)




		# First sprout
		self.assertTrue(os.path.isfile(self.f_mss1_md_rock))
		self.assertTrue(os.path.isfile(self.f_mss1_md_paper))
		self.assertTrue(os.path.isfile(self.f_mss1_md_scissors))
		self.assertTrue(os.path.isfile(self.f_mss1_md_gc_sadmovies))
		self.assertTrue(os.path.isfile(self.f_mss1_md_gc_dumbasses))
		self.assertTrue(os.path.isfile(self.f_mss1_md_gc_onion))
		self.assertTrue(os.path.isdir(self.d_mss1_c_good))
		self.assertTrue(os.path.isdir(self.d_mss1_c_notgood))
		self.assertTrue(os.path.isdir(self.d_mss1_c_bad))
		self.assertTrue(os.path.isdir(self.d_mss1_c_notbad))
		self.assertTrue(os.path.isdir(self.d_mss1_c_notnot))
		self.assertTrue(os.path.isdir(self.d_mss1_c_badbad))
		self.assertTrue(os.path.isdir(self.d_mss1_c_badgood))
		self.assertTrue(os.path.isdir(self.d_mss1_c_goodbad))

		# Second sprout
		self.assertTrue(os.path.isfile(self.f_mss2_tasksschedule))
		self.assertTrue(os.path.isfile(self.f_mss2_explangraph))
		self.assertTrue(os.path.isfile(self.f_mss2_c_hardcleaning))
		self.assertTrue(os.path.isfile(self.f_mss2_c_hardwashing))
		self.assertTrue(os.path.isfile(self.f_mss2_c_mediumcooking))
		self.assertTrue(os.path.isfile(self.f_mss2_c_mediumhomework))
		self.assertTrue(os.path.isfile(self.f_mss2_c_mediumphysicals))
		self.assertTrue(os.path.isfile(self.f_mss2_c_easyprocrastinating))
		self.assertTrue(os.path.isfile(self.f_mss2_c_easyreflections))
		self.assertTrue(os.path.isfile(self.f_mss2_c_easyphilosophizing))

		# Third sprout
		self.assertTrue(os.path.isfile(self.f_mss3_k_f_milk))
		self.assertTrue(os.path.isfile(self.f_mss3_k_f_cheese))
		self.assertTrue(os.path.isfile(self.f_mss3_k_f_meat))
		self.assertTrue(os.path.isfile(self.f_mss3_k_o_chicken))
		self.assertTrue(os.path.isfile(self.f_mss3_k_o_pie))
		self.assertTrue(os.path.isfile(self.f_mss3_k_t_bc_bread))
		self.assertTrue(os.path.isfile(self.f_mss3_k_t_bc_crumb1))
		self.assertTrue(os.path.isfile(self.f_mss3_k_t_bc_crumb2))
		self.assertTrue(os.path.isfile(self.f_mss3_k_t_bc_crumb420))
		self.assertTrue(os.path.isfile(self.f_mss3_k_t_bc_crumb69))




		# First sprout first bough
		self.assertTrue(os.path.isdir(self.tg1_d_mss1_md))
		self.assertTrue(os.path.isdir(self.tg1_d_mss1_md_gc))
		self.assertTrue(os.path.isdir(self.tg1_d_mss1_c))
		self.assertTrue(os.path.isdir(self.tg1_d_mss1_c_good))
		self.assertTrue(os.path.isdir(self.tg1_d_mss1_c_notgood))
		self.assertFalse(os.path.isdir(self.tg1_d_mss1_c_bad))
		self.assertFalse(os.path.isdir(self.tg1_d_mss1_c_notbad))
		self.assertTrue(os.path.isdir(self.tg1_d_mss1_c_notnot))
		self.assertFalse(os.path.isdir(self.tg1_d_mss1_c_badbad))
		self.assertFalse(os.path.isdir(self.tg1_d_mss1_c_badgood))
		self.assertFalse(os.path.isdir(self.tg1_d_mss1_c_goodbad))
		self.assertTrue(os.path.isfile(self.tg1_f_mss1_md_rock))
		self.assertTrue(os.path.isfile(self.tg1_f_mss1_md_paper))
		self.assertTrue(os.path.isfile(self.tg1_f_mss1_md_scissors))
		self.assertTrue(os.path.isfile(self.tg1_f_mss1_md_gc_sadmovies))
		self.assertTrue(os.path.isfile(self.tg1_f_mss1_md_gc_dumbasses))
		self.assertTrue(os.path.isfile(self.tg1_f_mss1_md_gc_onion))

		# Second sprout first bough
		self.assertTrue(os.path.isdir(self.tg1_d_mss2_c))
		self.assertFalse(os.path.isfile(self.tg1_f_mss2_tasksschedule))
		self.assertFalse(os.path.isfile(self.tg1_f_mss2_explangraph))
		self.assertFalse(os.path.isfile(self.tg1_f_mss2_c_hardcleaning))
		self.assertFalse(os.path.isfile(self.tg1_f_mss2_c_hardwashing))
		self.assertFalse(os.path.isfile(self.tg1_f_mss2_c_mediumcooking))
		self.assertFalse(os.path.isfile(self.tg1_f_mss2_c_mediumhomework))
		self.assertFalse(os.path.isfile(self.tg1_f_mss2_c_mediumphysicals))
		self.assertFalse(os.path.isfile(self.tg1_f_mss2_c_easyprocrastinating))
		self.assertFalse(os.path.isfile(self.tg1_f_mss2_c_easyreflections))
		self.assertFalse(os.path.isfile(self.tg1_f_mss2_c_easyphilosophizing))

		# Third sprout first bough
		self.assertTrue(os.path.isdir(self.tg1_d_mss3_k))
		self.assertTrue(os.path.isdir(self.tg1_d_mss3_k_f))
		self.assertTrue(os.path.isdir(self.tg1_d_mss3_k_o))
		self.assertTrue(os.path.isdir(self.tg1_d_mss3_k_t))
		self.assertTrue(os.path.isdir(self.tg1_d_mss3_k_t_bc))
		self.assertTrue(os.path.isfile(self.tg1_f_mss3_k_f_milk))
		self.assertTrue(os.path.isfile(self.tg1_f_mss3_k_f_cheese))
		self.assertTrue(os.path.isfile(self.tg1_f_mss3_k_f_meat))
		self.assertTrue(os.path.isfile(self.tg1_f_mss3_k_o_chicken))
		self.assertTrue(os.path.isfile(self.tg1_f_mss3_k_o_pie))
		self.assertTrue(os.path.isfile(self.tg1_f_mss3_k_t_bc_bread))
		self.assertFalse(os.path.isfile(self.tg1_f_mss3_k_t_bc_crumb1))
		self.assertFalse(os.path.isfile(self.tg1_f_mss3_k_t_bc_crumb2))
		self.assertFalse(os.path.isfile(self.tg1_f_mss3_k_t_bc_crumb420))
		self.assertFalse(os.path.isfile(self.tg1_f_mss3_k_t_bc_crumb69))
















	def test_no_touch_flourish(self):

		sleep(1.1)
		self.test_case.loggy.init_name = "no_touch_flourish"


		# First sprout
		self.assertTrue(os.path.isfile(self.f_mss1_md_rock))
		self.assertTrue(os.path.isfile(self.f_mss1_md_paper))
		self.assertTrue(os.path.isfile(self.f_mss1_md_scissors))
		self.assertTrue(os.path.isfile(self.f_mss1_md_gc_sadmovies))
		self.assertTrue(os.path.isfile(self.f_mss1_md_gc_dumbasses))
		self.assertTrue(os.path.isfile(self.f_mss1_md_gc_onion))
		self.assertTrue(os.path.isdir(self.d_mss1_c_good))
		self.assertTrue(os.path.isdir(self.d_mss1_c_notgood))
		self.assertTrue(os.path.isdir(self.d_mss1_c_bad))
		self.assertTrue(os.path.isdir(self.d_mss1_c_notbad))
		self.assertTrue(os.path.isdir(self.d_mss1_c_notnot))
		self.assertTrue(os.path.isdir(self.d_mss1_c_badbad))
		self.assertTrue(os.path.isdir(self.d_mss1_c_badgood))
		self.assertTrue(os.path.isdir(self.d_mss1_c_goodbad))

		# Second sprout
		self.assertTrue(os.path.isfile(self.f_mss2_tasksschedule))
		self.assertTrue(os.path.isfile(self.f_mss2_explangraph))
		self.assertTrue(os.path.isfile(self.f_mss2_c_hardcleaning))
		self.assertTrue(os.path.isfile(self.f_mss2_c_hardwashing))
		self.assertTrue(os.path.isfile(self.f_mss2_c_mediumcooking))
		self.assertTrue(os.path.isfile(self.f_mss2_c_mediumhomework))
		self.assertTrue(os.path.isfile(self.f_mss2_c_mediumphysicals))
		self.assertTrue(os.path.isfile(self.f_mss2_c_easyprocrastinating))
		self.assertTrue(os.path.isfile(self.f_mss2_c_easyreflections))
		self.assertTrue(os.path.isfile(self.f_mss2_c_easyphilosophizing))

		# Third sprout
		self.assertTrue(os.path.isfile(self.f_mss3_k_f_milk))
		self.assertTrue(os.path.isfile(self.f_mss3_k_f_cheese))
		self.assertTrue(os.path.isfile(self.f_mss3_k_f_meat))
		self.assertTrue(os.path.isfile(self.f_mss3_k_o_chicken))
		self.assertTrue(os.path.isfile(self.f_mss3_k_o_pie))
		self.assertTrue(os.path.isfile(self.f_mss3_k_t_bc_bread))
		self.assertTrue(os.path.isfile(self.f_mss3_k_t_bc_crumb1))
		self.assertTrue(os.path.isfile(self.f_mss3_k_t_bc_crumb2))
		self.assertTrue(os.path.isfile(self.f_mss3_k_t_bc_crumb420))
		self.assertTrue(os.path.isfile(self.f_mss3_k_t_bc_crumb69))




		with self.assertLogs("no_touch_flourish", 10) as case_loggy:

			self.test_case = self.test_case()
			self.test_case.sync()




		self.no_loggy_levels(case_loggy.output, 30,40,50)
		self.assertIn(

			f"DEBUG:no_touch_flourish:New draft peek \"{self.f_mss1_md_rock}\" "
			f"new ring: {os.path.getmtime(self.f_mss1_md_rock)}",
			case_loggy.output
		)
		self.assertIn(

			f"DEBUG:no_touch_flourish:New draft peek \"{self.f_mss1_md_paper}\" "
			f"new ring: {os.path.getmtime(self.f_mss1_md_paper)}",
			case_loggy.output
		)
		self.assertIn(

			f"DEBUG:no_touch_flourish:New draft peek \"{self.f_mss1_md_scissors}\" "
			f"new ring: {os.path.getmtime(self.f_mss1_md_scissors)}",
			case_loggy.output
		)
		self.assertIn(

			f"DEBUG:no_touch_flourish:New draft peek \"{self.f_mss1_md_gc_sadmovies}\" "
			f"new ring: {os.path.getmtime(self.f_mss1_md_gc_sadmovies)}",
			case_loggy.output
		)
		self.assertIn(

			f"DEBUG:no_touch_flourish:New draft peek \"{self.f_mss1_md_gc_dumbasses}\" "
			f"new ring: {os.path.getmtime(self.f_mss1_md_gc_dumbasses)}",
			case_loggy.output
		)
		self.assertIn(

			f"DEBUG:no_touch_flourish:New draft peek \"{self.f_mss1_md_gc_onion}\" "
			f"new ring: {os.path.getmtime(self.f_mss1_md_gc_onion)}",
			case_loggy.output
		)




		self.assertNotIn(

			f"DEBUG:no_touch_flourish:New draft peek \"{self.f_mss2_tasksschedule}\" "
			f"new ring: {os.path.getmtime(self.f_mss2_tasksschedule)}",
			case_loggy.output
		)
		self.assertNotIn(

			f"DEBUG:no_touch_flourish:New draft peek \"{self.f_mss2_explangraph}\" "
			f"new ring: {os.path.getmtime(self.f_mss2_explangraph)}",
			case_loggy.output
		)
		self.assertNotIn(

			f"DEBUG:no_touch_flourish:New draft peek \"{self.f_mss2_c_hardcleaning}\" "
			f"new ring: {os.path.getmtime(self.f_mss2_c_hardcleaning)}",
			case_loggy.output
		)
		self.assertNotIn(

			f"DEBUG:no_touch_flourish:New draft peek \"{self.f_mss2_c_hardwashing}\" "
			f"new ring: {os.path.getmtime(self.f_mss2_c_hardwashing)}",
			case_loggy.output
		)
		self.assertNotIn(

			f"DEBUG:no_touch_flourish:New draft peek \"{self.f_mss2_c_mediumcooking}\" "
			f"new ring: {os.path.getmtime(self.f_mss2_c_mediumcooking)}",
			case_loggy.output
		)
		self.assertNotIn(

			f"DEBUG:no_touch_flourish:New draft peek \"{self.f_mss2_c_mediumhomework}\" "
			f"new ring: {os.path.getmtime(self.f_mss2_c_mediumhomework)}",
			case_loggy.output
		)
		self.assertNotIn(

			f"DEBUG:no_touch_flourish:New draft peek \"{self.f_mss2_c_mediumphysicals}\" "
			f"new ring: {os.path.getmtime(self.f_mss2_c_mediumphysicals)}",
			case_loggy.output
		)
		self.assertNotIn(

			f"DEBUG:no_touch_flourish:New draft peek \"{self.f_mss2_c_easyprocrastinating}\" "
			f"new ring: {os.path.getmtime(self.f_mss2_c_easyprocrastinating)}",
			case_loggy.output
		)
		self.assertNotIn(

			f"DEBUG:no_touch_flourish:New draft peek \"{self.f_mss2_c_easyreflections}\" "
			f"new ring: {os.path.getmtime(self.f_mss2_c_easyreflections)}",
			case_loggy.output
		)
		self.assertNotIn(

			f"DEBUG:no_touch_flourish:New draft peek \"{self.f_mss2_c_easyphilosophizing}\" "
			f"new ring: {os.path.getmtime(self.f_mss2_c_easyphilosophizing)}",
			case_loggy.output
		)




		self.assertIn(

			f"DEBUG:no_touch_flourish:New draft peek \"{self.f_mss3_k_f_milk}\" "
			f"new ring: {os.path.getmtime(self.f_mss3_k_f_milk)}",
			case_loggy.output
		)
		self.assertIn(

			f"DEBUG:no_touch_flourish:New draft peek \"{self.f_mss3_k_f_cheese}\" "
			f"new ring: {os.path.getmtime(self.f_mss3_k_f_cheese)}",
			case_loggy.output
		)
		self.assertIn(

			f"DEBUG:no_touch_flourish:New draft peek \"{self.f_mss3_k_f_meat}\" "
			f"new ring: {os.path.getmtime(self.f_mss3_k_f_meat)}",
			case_loggy.output
		)
		self.assertIn(

			f"DEBUG:no_touch_flourish:New draft peek \"{self.f_mss3_k_o_chicken}\" "
			f"new ring: {os.path.getmtime(self.f_mss3_k_o_chicken)}",
			case_loggy.output
		)
		self.assertIn(

			f"DEBUG:no_touch_flourish:New draft peek \"{self.f_mss3_k_o_pie}\" "
			f"new ring: {os.path.getmtime(self.f_mss3_k_o_pie)}",
			case_loggy.output
		)
		self.assertIn(

			f"DEBUG:no_touch_flourish:New draft peek \"{self.f_mss3_k_t_bc_bread}\" "
			f"new ring: {os.path.getmtime(self.f_mss3_k_t_bc_bread)}",
			case_loggy.output
		)
		self.assertNotIn(

			f"DEBUG:no_touch_flourish:New draft peek \"{self.f_mss3_k_t_bc_crumb1}\" "
			f"new ring: {os.path.getmtime(self.f_mss3_k_t_bc_crumb1)}",
			case_loggy.output
		)
		self.assertNotIn(

			f"DEBUG:no_touch_flourish:New draft peek \"{self.f_mss3_k_t_bc_crumb2}\" "
			f"new ring: {os.path.getmtime(self.f_mss3_k_t_bc_crumb2)}",
			case_loggy.output
		)
		self.assertNotIn(

			f"DEBUG:no_touch_flourish:New draft peek \"{self.f_mss3_k_t_bc_crumb420}\" "
			f"new ring: {os.path.getmtime(self.f_mss3_k_t_bc_crumb420)}",
			case_loggy.output
		)
		self.assertNotIn(

			f"DEBUG:no_touch_flourish:New draft peek \"{self.f_mss3_k_t_bc_crumb69}\" "
			f"new ring: {os.path.getmtime(self.f_mss3_k_t_bc_crumb69)}",
			case_loggy.output
		)




		# First sprout
		self.assertTrue(os.path.isfile(self.f_mss1_md_rock))
		self.assertTrue(os.path.isfile(self.f_mss1_md_paper))
		self.assertTrue(os.path.isfile(self.f_mss1_md_scissors))
		self.assertTrue(os.path.isfile(self.f_mss1_md_gc_sadmovies))
		self.assertTrue(os.path.isfile(self.f_mss1_md_gc_dumbasses))
		self.assertTrue(os.path.isfile(self.f_mss1_md_gc_onion))
		self.assertTrue(os.path.isdir(self.d_mss1_c_good))
		self.assertTrue(os.path.isdir(self.d_mss1_c_notgood))
		self.assertTrue(os.path.isdir(self.d_mss1_c_bad))
		self.assertTrue(os.path.isdir(self.d_mss1_c_notbad))
		self.assertTrue(os.path.isdir(self.d_mss1_c_notnot))
		self.assertTrue(os.path.isdir(self.d_mss1_c_badbad))
		self.assertTrue(os.path.isdir(self.d_mss1_c_badgood))
		self.assertTrue(os.path.isdir(self.d_mss1_c_goodbad))

		# Second sprout
		self.assertTrue(os.path.isfile(self.f_mss2_tasksschedule))
		self.assertTrue(os.path.isfile(self.f_mss2_explangraph))
		self.assertTrue(os.path.isfile(self.f_mss2_c_hardcleaning))
		self.assertTrue(os.path.isfile(self.f_mss2_c_hardwashing))
		self.assertTrue(os.path.isfile(self.f_mss2_c_mediumcooking))
		self.assertTrue(os.path.isfile(self.f_mss2_c_mediumhomework))
		self.assertTrue(os.path.isfile(self.f_mss2_c_mediumphysicals))
		self.assertTrue(os.path.isfile(self.f_mss2_c_easyprocrastinating))
		self.assertTrue(os.path.isfile(self.f_mss2_c_easyreflections))
		self.assertTrue(os.path.isfile(self.f_mss2_c_easyphilosophizing))

		# Third sprout
		self.assertTrue(os.path.isfile(self.f_mss3_k_f_milk))
		self.assertTrue(os.path.isfile(self.f_mss3_k_f_cheese))
		self.assertTrue(os.path.isfile(self.f_mss3_k_f_meat))
		self.assertTrue(os.path.isfile(self.f_mss3_k_o_chicken))
		self.assertTrue(os.path.isfile(self.f_mss3_k_o_pie))
		self.assertTrue(os.path.isfile(self.f_mss3_k_t_bc_bread))
		self.assertTrue(os.path.isfile(self.f_mss3_k_t_bc_crumb1))
		self.assertTrue(os.path.isfile(self.f_mss3_k_t_bc_crumb2))
		self.assertTrue(os.path.isfile(self.f_mss3_k_t_bc_crumb420))
		self.assertTrue(os.path.isfile(self.f_mss3_k_t_bc_crumb69))




		# First sprout first bough
		self.assertTrue(os.path.isdir(self.tg1_d_mss1_md))
		self.assertTrue(os.path.isdir(self.tg1_d_mss1_md_gc))
		self.assertTrue(os.path.isdir(self.tg1_d_mss1_c))
		self.assertTrue(os.path.isdir(self.tg1_d_mss1_c_good))
		self.assertTrue(os.path.isdir(self.tg1_d_mss1_c_notgood))
		self.assertFalse(os.path.isdir(self.tg1_d_mss1_c_bad))
		self.assertFalse(os.path.isdir(self.tg1_d_mss1_c_notbad))
		self.assertTrue(os.path.isdir(self.tg1_d_mss1_c_notnot))
		self.assertFalse(os.path.isdir(self.tg1_d_mss1_c_badbad))
		self.assertFalse(os.path.isdir(self.tg1_d_mss1_c_badgood))
		self.assertFalse(os.path.isdir(self.tg1_d_mss1_c_goodbad))
		self.assertTrue(os.path.isfile(self.tg1_f_mss1_md_rock))
		self.assertTrue(os.path.isfile(self.tg1_f_mss1_md_paper))
		self.assertTrue(os.path.isfile(self.tg1_f_mss1_md_scissors))
		self.assertTrue(os.path.isfile(self.tg1_f_mss1_md_gc_sadmovies))
		self.assertTrue(os.path.isfile(self.tg1_f_mss1_md_gc_dumbasses))
		self.assertTrue(os.path.isfile(self.tg1_f_mss1_md_gc_onion))

		# Second sprout first bough
		self.assertTrue(os.path.isdir(self.tg1_d_mss2_c))
		self.assertFalse(os.path.isfile(self.tg1_f_mss2_tasksschedule))
		self.assertFalse(os.path.isfile(self.tg1_f_mss2_explangraph))
		self.assertFalse(os.path.isfile(self.tg1_f_mss2_c_hardcleaning))
		self.assertFalse(os.path.isfile(self.tg1_f_mss2_c_hardwashing))
		self.assertFalse(os.path.isfile(self.tg1_f_mss2_c_mediumcooking))
		self.assertFalse(os.path.isfile(self.tg1_f_mss2_c_mediumhomework))
		self.assertFalse(os.path.isfile(self.tg1_f_mss2_c_mediumphysicals))
		self.assertFalse(os.path.isfile(self.tg1_f_mss2_c_easyprocrastinating))
		self.assertFalse(os.path.isfile(self.tg1_f_mss2_c_easyreflections))
		self.assertFalse(os.path.isfile(self.tg1_f_mss2_c_easyphilosophizing))

		# Third sprout first bough
		self.assertTrue(os.path.isdir(self.tg1_d_mss3_k))
		self.assertTrue(os.path.isdir(self.tg1_d_mss3_k_f))
		self.assertTrue(os.path.isdir(self.tg1_d_mss3_k_o))
		self.assertTrue(os.path.isdir(self.tg1_d_mss3_k_t))
		self.assertTrue(os.path.isdir(self.tg1_d_mss3_k_t_bc))
		self.assertTrue(os.path.isfile(self.tg1_f_mss3_k_f_milk))
		self.assertTrue(os.path.isfile(self.tg1_f_mss3_k_f_cheese))
		self.assertTrue(os.path.isfile(self.tg1_f_mss3_k_f_meat))
		self.assertTrue(os.path.isfile(self.tg1_f_mss3_k_o_chicken))
		self.assertTrue(os.path.isfile(self.tg1_f_mss3_k_o_pie))
		self.assertTrue(os.path.isfile(self.tg1_f_mss3_k_t_bc_bread))
		self.assertFalse(os.path.isfile(self.tg1_f_mss3_k_t_bc_crumb1))
		self.assertFalse(os.path.isfile(self.tg1_f_mss3_k_t_bc_crumb2))
		self.assertFalse(os.path.isfile(self.tg1_f_mss3_k_t_bc_crumb420))
		self.assertFalse(os.path.isfile(self.tg1_f_mss3_k_t_bc_crumb69))
















	def test_touch_flourish(self):

		sleep(1.1)
		self.test_case.loggy.init_name = "touch_flourish"
		self.fmake(self.f_mss1_md_gc_onion, "makes even devil cry")
		self.fmake(self.f_mss2_tasksschedule, "1. get up\n2. get sad")
		self.fmake(self.f_mss3_k_t_bc_bread, "pit")


		# First sprout
		self.assertTrue(os.path.isfile(self.f_mss1_md_rock))
		self.assertTrue(os.path.isfile(self.f_mss1_md_paper))
		self.assertTrue(os.path.isfile(self.f_mss1_md_scissors))
		self.assertTrue(os.path.isfile(self.f_mss1_md_gc_sadmovies))
		self.assertTrue(os.path.isfile(self.f_mss1_md_gc_dumbasses))
		self.assertTrue(os.path.isfile(self.f_mss1_md_gc_onion))
		self.assertTrue(os.path.isdir(self.d_mss1_c_good))
		self.assertTrue(os.path.isdir(self.d_mss1_c_notgood))
		self.assertTrue(os.path.isdir(self.d_mss1_c_bad))
		self.assertTrue(os.path.isdir(self.d_mss1_c_notbad))
		self.assertTrue(os.path.isdir(self.d_mss1_c_notnot))
		self.assertTrue(os.path.isdir(self.d_mss1_c_badbad))
		self.assertTrue(os.path.isdir(self.d_mss1_c_badgood))
		self.assertTrue(os.path.isdir(self.d_mss1_c_goodbad))

		# Second sprout
		self.assertTrue(os.path.isfile(self.f_mss2_tasksschedule))
		self.assertTrue(os.path.isfile(self.f_mss2_explangraph))
		self.assertTrue(os.path.isfile(self.f_mss2_c_hardcleaning))
		self.assertTrue(os.path.isfile(self.f_mss2_c_hardwashing))
		self.assertTrue(os.path.isfile(self.f_mss2_c_mediumcooking))
		self.assertTrue(os.path.isfile(self.f_mss2_c_mediumhomework))
		self.assertTrue(os.path.isfile(self.f_mss2_c_mediumphysicals))
		self.assertTrue(os.path.isfile(self.f_mss2_c_easyprocrastinating))
		self.assertTrue(os.path.isfile(self.f_mss2_c_easyreflections))
		self.assertTrue(os.path.isfile(self.f_mss2_c_easyphilosophizing))

		# Third sprout
		self.assertTrue(os.path.isfile(self.f_mss3_k_f_milk))
		self.assertTrue(os.path.isfile(self.f_mss3_k_f_cheese))
		self.assertTrue(os.path.isfile(self.f_mss3_k_f_meat))
		self.assertTrue(os.path.isfile(self.f_mss3_k_o_chicken))
		self.assertTrue(os.path.isfile(self.f_mss3_k_o_pie))
		self.assertTrue(os.path.isfile(self.f_mss3_k_t_bc_bread))
		self.assertTrue(os.path.isfile(self.f_mss3_k_t_bc_crumb1))
		self.assertTrue(os.path.isfile(self.f_mss3_k_t_bc_crumb2))
		self.assertTrue(os.path.isfile(self.f_mss3_k_t_bc_crumb420))
		self.assertTrue(os.path.isfile(self.f_mss3_k_t_bc_crumb69))




		with self.assertLogs("touch_flourish", 10) as case_loggy:

			self.test_case = self.test_case()
			self.test_case.sync()




		self.no_loggy_levels(case_loggy.output, 30,40,50)
		self.assertIn(

			f"DEBUG:touch_flourish:New draft peek \"{self.f_mss1_md_rock}\" "
			f"new ring: {os.path.getmtime(self.f_mss1_md_rock)}",
			case_loggy.output
		)
		self.assertIn(

			f"DEBUG:touch_flourish:New draft peek \"{self.f_mss1_md_paper}\" "
			f"new ring: {os.path.getmtime(self.f_mss1_md_paper)}",
			case_loggy.output
		)
		self.assertIn(

			f"DEBUG:touch_flourish:New draft peek \"{self.f_mss1_md_scissors}\" "
			f"new ring: {os.path.getmtime(self.f_mss1_md_scissors)}",
			case_loggy.output
		)
		self.assertIn(

			f"DEBUG:touch_flourish:New draft peek \"{self.f_mss1_md_gc_sadmovies}\" "
			f"new ring: {os.path.getmtime(self.f_mss1_md_gc_sadmovies)}",
			case_loggy.output
		)
		self.assertIn(

			f"DEBUG:touch_flourish:New draft peek \"{self.f_mss1_md_gc_dumbasses}\" "
			f"new ring: {os.path.getmtime(self.f_mss1_md_gc_dumbasses)}",
			case_loggy.output
		)
		self.assertIn(

			f"DEBUG:touch_flourish:New draft peek \"{self.f_mss1_md_gc_onion}\" "
			f"new ring: {os.path.getmtime(self.f_mss1_md_gc_onion)}",
			case_loggy.output
		)




		self.assertNotIn(

			f"DEBUG:touch_flourish:New draft peek \"{self.f_mss2_tasksschedule}\" "
			f"new ring: {os.path.getmtime(self.f_mss2_tasksschedule)}",
			case_loggy.output
		)
		self.assertNotIn(

			f"DEBUG:touch_flourish:New draft peek \"{self.f_mss2_explangraph}\" "
			f"new ring: {os.path.getmtime(self.f_mss2_explangraph)}",
			case_loggy.output
		)
		self.assertNotIn(

			f"DEBUG:touch_flourish:New draft peek \"{self.f_mss2_c_hardcleaning}\" "
			f"new ring: {os.path.getmtime(self.f_mss2_c_hardcleaning)}",
			case_loggy.output
		)
		self.assertNotIn(

			f"DEBUG:touch_flourish:New draft peek \"{self.f_mss2_c_hardwashing}\" "
			f"new ring: {os.path.getmtime(self.f_mss2_c_hardwashing)}",
			case_loggy.output
		)
		self.assertNotIn(

			f"DEBUG:touch_flourish:New draft peek \"{self.f_mss2_c_mediumcooking}\" "
			f"new ring: {os.path.getmtime(self.f_mss2_c_mediumcooking)}",
			case_loggy.output
		)
		self.assertNotIn(

			f"DEBUG:touch_flourish:New draft peek \"{self.f_mss2_c_mediumhomework}\" "
			f"new ring: {os.path.getmtime(self.f_mss2_c_mediumhomework)}",
			case_loggy.output
		)
		self.assertNotIn(

			f"DEBUG:touch_flourish:New draft peek \"{self.f_mss2_c_mediumphysicals}\" "
			f"new ring: {os.path.getmtime(self.f_mss2_c_mediumphysicals)}",
			case_loggy.output
		)
		self.assertNotIn(

			f"DEBUG:touch_flourish:New draft peek \"{self.f_mss2_c_easyprocrastinating}\" "
			f"new ring: {os.path.getmtime(self.f_mss2_c_easyprocrastinating)}",
			case_loggy.output
		)
		self.assertNotIn(

			f"DEBUG:touch_flourish:New draft peek \"{self.f_mss2_c_easyreflections}\" "
			f"new ring: {os.path.getmtime(self.f_mss2_c_easyreflections)}",
			case_loggy.output
		)
		self.assertNotIn(

			f"DEBUG:touch_flourish:New draft peek \"{self.f_mss2_c_easyphilosophizing}\" "
			f"new ring: {os.path.getmtime(self.f_mss2_c_easyphilosophizing)}",
			case_loggy.output
		)




		self.assertIn(

			f"DEBUG:touch_flourish:New draft peek \"{self.f_mss3_k_f_milk}\" "
			f"new ring: {os.path.getmtime(self.f_mss3_k_f_milk)}",
			case_loggy.output
		)
		self.assertIn(

			f"DEBUG:touch_flourish:New draft peek \"{self.f_mss3_k_f_cheese}\" "
			f"new ring: {os.path.getmtime(self.f_mss3_k_f_cheese)}",
			case_loggy.output
		)
		self.assertIn(

			f"DEBUG:touch_flourish:New draft peek \"{self.f_mss3_k_f_meat}\" "
			f"new ring: {os.path.getmtime(self.f_mss3_k_f_meat)}",
			case_loggy.output
		)
		self.assertIn(

			f"DEBUG:touch_flourish:New draft peek \"{self.f_mss3_k_o_chicken}\" "
			f"new ring: {os.path.getmtime(self.f_mss3_k_o_chicken)}",
			case_loggy.output
		)
		self.assertIn(

			f"DEBUG:touch_flourish:New draft peek \"{self.f_mss3_k_o_pie}\" "
			f"new ring: {os.path.getmtime(self.f_mss3_k_o_pie)}",
			case_loggy.output
		)
		self.assertIn(

			f"DEBUG:touch_flourish:New draft peek \"{self.f_mss3_k_t_bc_bread}\" "
			f"new ring: {os.path.getmtime(self.f_mss3_k_t_bc_bread)}",
			case_loggy.output
		)
		self.assertNotIn(

			f"DEBUG:touch_flourish:New draft peek \"{self.f_mss3_k_t_bc_crumb1}\" "
			f"new ring: {os.path.getmtime(self.f_mss3_k_t_bc_crumb1)}",
			case_loggy.output
		)
		self.assertNotIn(

			f"DEBUG:touch_flourish:New draft peek \"{self.f_mss3_k_t_bc_crumb2}\" "
			f"new ring: {os.path.getmtime(self.f_mss3_k_t_bc_crumb2)}",
			case_loggy.output
		)
		self.assertNotIn(

			f"DEBUG:touch_flourish:New draft peek \"{self.f_mss3_k_t_bc_crumb420}\" "
			f"new ring: {os.path.getmtime(self.f_mss3_k_t_bc_crumb420)}",
			case_loggy.output
		)
		self.assertNotIn(

			f"DEBUG:touch_flourish:New draft peek \"{self.f_mss3_k_t_bc_crumb69}\" "
			f"new ring: {os.path.getmtime(self.f_mss3_k_t_bc_crumb69)}",
			case_loggy.output
		)




		# First sprout
		self.assertTrue(os.path.isfile(self.f_mss1_md_rock))
		self.assertTrue(os.path.isfile(self.f_mss1_md_paper))
		self.assertTrue(os.path.isfile(self.f_mss1_md_scissors))
		self.assertTrue(os.path.isfile(self.f_mss1_md_gc_sadmovies))
		self.assertTrue(os.path.isfile(self.f_mss1_md_gc_dumbasses))
		self.assertTrue(os.path.isfile(self.f_mss1_md_gc_onion))
		self.assertTrue(os.path.isdir(self.d_mss1_c_good))
		self.assertTrue(os.path.isdir(self.d_mss1_c_notgood))
		self.assertTrue(os.path.isdir(self.d_mss1_c_bad))
		self.assertTrue(os.path.isdir(self.d_mss1_c_notbad))
		self.assertTrue(os.path.isdir(self.d_mss1_c_notnot))
		self.assertTrue(os.path.isdir(self.d_mss1_c_badbad))
		self.assertTrue(os.path.isdir(self.d_mss1_c_badgood))
		self.assertTrue(os.path.isdir(self.d_mss1_c_goodbad))

		# Second sprout
		self.assertTrue(os.path.isfile(self.f_mss2_tasksschedule))
		self.assertTrue(os.path.isfile(self.f_mss2_explangraph))
		self.assertTrue(os.path.isfile(self.f_mss2_c_hardcleaning))
		self.assertTrue(os.path.isfile(self.f_mss2_c_hardwashing))
		self.assertTrue(os.path.isfile(self.f_mss2_c_mediumcooking))
		self.assertTrue(os.path.isfile(self.f_mss2_c_mediumhomework))
		self.assertTrue(os.path.isfile(self.f_mss2_c_mediumphysicals))
		self.assertTrue(os.path.isfile(self.f_mss2_c_easyprocrastinating))
		self.assertTrue(os.path.isfile(self.f_mss2_c_easyreflections))
		self.assertTrue(os.path.isfile(self.f_mss2_c_easyphilosophizing))

		# Third sprout
		self.assertTrue(os.path.isfile(self.f_mss3_k_f_milk))
		self.assertTrue(os.path.isfile(self.f_mss3_k_f_cheese))
		self.assertTrue(os.path.isfile(self.f_mss3_k_f_meat))
		self.assertTrue(os.path.isfile(self.f_mss3_k_o_chicken))
		self.assertTrue(os.path.isfile(self.f_mss3_k_o_pie))
		self.assertTrue(os.path.isfile(self.f_mss3_k_t_bc_bread))
		self.assertTrue(os.path.isfile(self.f_mss3_k_t_bc_crumb1))
		self.assertTrue(os.path.isfile(self.f_mss3_k_t_bc_crumb2))
		self.assertTrue(os.path.isfile(self.f_mss3_k_t_bc_crumb420))
		self.assertTrue(os.path.isfile(self.f_mss3_k_t_bc_crumb69))




		# First sprout first bough
		self.assertTrue(os.path.isdir(self.tg1_d_mss1_md))
		self.assertTrue(os.path.isdir(self.tg1_d_mss1_md_gc))
		self.assertTrue(os.path.isdir(self.tg1_d_mss1_c))
		self.assertTrue(os.path.isdir(self.tg1_d_mss1_c_good))
		self.assertTrue(os.path.isdir(self.tg1_d_mss1_c_notgood))
		self.assertFalse(os.path.isdir(self.tg1_d_mss1_c_bad))
		self.assertFalse(os.path.isdir(self.tg1_d_mss1_c_notbad))
		self.assertTrue(os.path.isdir(self.tg1_d_mss1_c_notnot))
		self.assertFalse(os.path.isdir(self.tg1_d_mss1_c_badbad))
		self.assertFalse(os.path.isdir(self.tg1_d_mss1_c_badgood))
		self.assertFalse(os.path.isdir(self.tg1_d_mss1_c_goodbad))
		self.assertTrue(os.path.isfile(self.tg1_f_mss1_md_rock))
		self.assertTrue(os.path.isfile(self.tg1_f_mss1_md_paper))
		self.assertTrue(os.path.isfile(self.tg1_f_mss1_md_scissors))
		self.assertTrue(os.path.isfile(self.tg1_f_mss1_md_gc_sadmovies))
		self.assertTrue(os.path.isfile(self.tg1_f_mss1_md_gc_dumbasses))
		self.assertTrue(os.path.isfile(self.tg1_f_mss1_md_gc_onion))

		# Second sprout first bough
		self.assertTrue(os.path.isdir(self.tg1_d_mss2_c))
		self.assertFalse(os.path.isfile(self.tg1_f_mss2_tasksschedule))
		self.assertFalse(os.path.isfile(self.tg1_f_mss2_explangraph))
		self.assertFalse(os.path.isfile(self.tg1_f_mss2_c_hardcleaning))
		self.assertFalse(os.path.isfile(self.tg1_f_mss2_c_hardwashing))
		self.assertFalse(os.path.isfile(self.tg1_f_mss2_c_mediumcooking))
		self.assertFalse(os.path.isfile(self.tg1_f_mss2_c_mediumhomework))
		self.assertFalse(os.path.isfile(self.tg1_f_mss2_c_mediumphysicals))
		self.assertFalse(os.path.isfile(self.tg1_f_mss2_c_easyprocrastinating))
		self.assertFalse(os.path.isfile(self.tg1_f_mss2_c_easyreflections))
		self.assertFalse(os.path.isfile(self.tg1_f_mss2_c_easyphilosophizing))

		# Third sprout first bough
		self.assertTrue(os.path.isdir(self.tg1_d_mss3_k))
		self.assertTrue(os.path.isdir(self.tg1_d_mss3_k_f))
		self.assertTrue(os.path.isdir(self.tg1_d_mss3_k_o))
		self.assertTrue(os.path.isdir(self.tg1_d_mss3_k_t))
		self.assertTrue(os.path.isdir(self.tg1_d_mss3_k_t_bc))
		self.assertTrue(os.path.isfile(self.tg1_f_mss3_k_f_milk))
		self.assertTrue(os.path.isfile(self.tg1_f_mss3_k_f_cheese))
		self.assertTrue(os.path.isfile(self.tg1_f_mss3_k_f_meat))
		self.assertTrue(os.path.isfile(self.tg1_f_mss3_k_o_chicken))
		self.assertTrue(os.path.isfile(self.tg1_f_mss3_k_o_pie))
		self.assertTrue(os.path.isfile(self.tg1_f_mss3_k_t_bc_bread))
		self.assertFalse(os.path.isfile(self.tg1_f_mss3_k_t_bc_crumb1))
		self.assertFalse(os.path.isfile(self.tg1_f_mss3_k_t_bc_crumb2))
		self.assertFalse(os.path.isfile(self.tg1_f_mss3_k_t_bc_crumb420))
		self.assertFalse(os.path.isfile(self.tg1_f_mss3_k_t_bc_crumb69))








if	__name__ == "__main__" : unittest.main(verbosity=2)










from pathlib					import Path
from typing						import List
from pygwarts.magical.spells	import patronus
from pygwarts.hagrid.bloom		import Bloom
from pygwarts.hagrid.thrivables	import Tree








class Germination(Bloom):

	"""
		Implementation of Bloom object that processes only folders by "bloom-interface":
			accepts:
				origin	- current Tree object to be processed;
				sprout	- current source root folder as a string;
				branch	- current source working directory as a Path object;
				twigs	- list of Path objects as current source folders;
				leafs	- list of Path objects as current source files.
			return:
				None.
		The folders processing (twigs planting) for current Bloom object __call__ relies on mandatory
		"thrive" callable, which will be invoked for every single Path object folder in "twigs" list,
		which declaration must be reachable for current Tree object "origin" by escalation and which
		must implement "planting interface" to accept following arguments:
			origin	- Tree object that is currently planting, used especially for emitting logging messages;
			sprout	- string that represents the current source root;
			branch	- Path object that represents current source folder;
			twig	- Path object that represents current "planting" source folder;
			bough	- Path object that represents current "planting" destination parent folder.
		This Bloom object designed to be the dispatcher for folders creation operation and does not
		copy any folders with content. The idea that is base for "Germination" name and that laid in such
		folders creation is the replication of the fs tree structure. Any files operations must be handled
		by corresponding Bloom objects, which handles file "planting" operations.
		Any Exception raised by "thrive" might not be handled by "thrive", as originally designed, but
		will be handled by this Bloom object. In this case, after the Exception will be handled,
		the iteration over twigs list will be continued. The iteration will be stopped if either
		"origin" argument is not a Tree object, or "origin" is a Tree object but with invalid
		"bough" field, or the "thrive" implementation either not callable or not found.
	"""

	def __call__(self, origin :Tree, sprout :str, branch :Path, twigs :List[Path], leafs :List[Path]):


		if	isinstance(twigs, list) and len(twigs):
			if	isinstance(origin, Tree) and isinstance((bough := getattr(origin, "bough", None)), Path):
				self.loggy.debug(f"Commencing germination of {len(twigs)} \"{bough}\" twigs")


				# IMPORTANT: "thrive" to be invoked from "origin" to ensure right escalation, as "planting"
				# invocation must be issued not from Bloom object but from Tree object to be processed.
				if	callable(thrive := getattr(origin, "thrive", None)):
					for twig in twigs:


						try:

							# Actual "planting" invocation relies on callable "thrive", that escalated from
							# the current Tree object, to implement "planting interface". The chain of calls,
							# that leads to this moment assumes all arguments are valid, cause the original
							# design of "planting" object doesn't involve arguments validation. That means
							# only "origin" validation at the beginning guards possible Exceptions, that might
							# be raised in "thrive" because of arguments inconsistency, and which all won't be
							# handled by original "planting" design but will be caught back in current Bloom.
							thrive(origin, sprout, branch, twig, bough)
							self.mark_tree(str(origin), "thrive")
							self.loggy.debug(f"Germinated \"{twig}\"")


						except	Exception as E:

							self.loggy.warning(f"Failed to germinate twig \"{twig}\" due to {patronus(E)}")
				else:		self.announce_implement(str(origin), "thrive")
			else:			self.announce_bough(str(origin), "germinate")
		else:				self.loggy.debug(f"Branch \"{str(branch)}\" has no twigs to germinate")










from typing									import Any
from typing									import Callable
from pygwarts.magical.philosophers_stone	import Transmutable
from pygwarts.magical.spells				import tentaclifors








class Transmutation(Transmutable):

	"""
		Pygwarts core class, that serves as a super class for one layer decorators (no arguments).
		For Transmutable objects decoration means transmutation, because actual decoration here means
		taking decorated class and produce another Transmutable class. with some modifications. The
		name of such object must remain the same as original decorated class, and this is achieved
		in declaration of actual decorator. Every Transmutation subclass must consist of a callable
		"_mutable_chain_injection" that will accept the only argument - the decorated Transmutable class.
		It must declare a new Transmutable class that inherits from the argument class. This new class
		must encompass all desired "transmutations" and must be returned by "_mutable_chain_injection".
		The decorated entity might be either a Transmutable, or any other type which will be treated
		special way that will lead to an Exception raise for the upper layer. For a Transmutable, as
		it might be another Transmutation, there is another special field "_CHAIN_LAYER_HOOK", that
		serves as a trigger and buffer at the same time. As the process of decorating by a class is
		handled by decorator __init__, it accepts the only positional argument "mutable_layer", which
		represents the decorated entity. If "mutable_layer" was successfully hooked by "_MUTABLE_CHAIN_HOOK"
		it is passed as a super class argument to "_mutable_chain_injection", that must be declared for
		current decorator, and the result Transmutable will be stored in "_CHAIN_LAYER_HOOK". Thus for
		another Transmutation that is happen to be a decorator for such decorator, the first thing to
		happen is to check "_CHAIN_LAYER_HOOK", so the Transmutable from the last "transmutation" that
		will be found there will be once again, but now for another decorator that decorated last decorator,
		passed as argument to "_mutable_chain_injection" that must be declared for this decorator and putted
		to it's own "_CHAIN_LAYER_HOOK", so decorating of decorators might appear any number of times. The
		final phase for Transmutation is the upper layer Transmutable phase two, when such Transmutation,
		after described above "transmutations" of decorated layers, or it's better say the final Transmutable
		class, will be instantiated with it's upper layer as an argument. This process is handled by __call__
		method and it is only goes for the "_CHAIN_LAYER_HOOK" content, as it must store that final
		Transmutable class. The process of mutable chain initiation for Transmutation schematically:

												->		decorator1.__init__(
							@decorator1			->			decorator2.__init__(
							@decorator2			->				Transmutable
							Transmutable		->			)
												->		).__call__(upper_layer_link)

		After initiation the upper layer will have transmutated member with the name of the original
		decorated class. The behavior of such object will solely depend on decorators implementation,
		but the only rule to keep in mind is the calling order top-down again, so for __call__ schematically:

					 ->	decorator1:
		@decorator1	 ->		__call__:
		@decorator2	 ->			super().__call__() ->	decorator2:
		Transmutable ->										__call__:
					 ->											super().__call__() -> Transmutable: __call__
	"""


	_CHAIN_LAYER_HOOK	= True


	def __init__(self, mutable_layer :Transmutable):

		if	callable(getattr(self, "_mutable_chain_injection", None)):
			match (mutation := getattr(mutable_layer, "_CHAIN_LAYER_HOOK", False)):


				# This case considers decorated "mutable_layer" is another Transmutation that wrapped it's
				# "mutable_layer" Transmutable in it's "_CHAIN_LAYER_HOOK".
				case type() if hasattr(mutation, "_MUTABLE_CHAIN_HOOK"):

					self._CHAIN_LAYER_HOOK = self._mutable_chain_injection(mutable_layer._CHAIN_LAYER_HOOK)
					return


				# Following two cases in charge for Decorated "mutable_layer" is another Transmutation that
				# tried to wrap not Transmutable class and did set it's "_CHAIN_LAYER_HOOK" to None, so
				# during mutable chain initiation TypError will be raised; decorated "mutable_layer" must
				# be inspected further.
				case None:

					self._CHAIN_LAYER_HOOK = None
					return


				case _: pass


			# At this point decorated "mutable_layer" must be either a Transmutable or not. If it is
			# Transmutable, it might be either the final point where decoration was started, or it might
			# also be ControlledTransmutation decorator, which differs from Transmutation as it doesn't
			# have additional final step for mutable chain initiation only, but return by it's initiation
			# transmutated Transmutable class. So for Transmutation the "mutable_layer" trasnmutation
			# will be stored in it's "_CHAIN_LAYER_HOOK" for either it's __call__ or other decorator.
			if	isinstance(mutable_layer, type) and getattr(mutable_layer, "_MUTABLE_CHAIN_HOOK", None):

				self._CHAIN_LAYER_HOOK = self._mutable_chain_injection(mutable_layer)
				return


		# This final point ensures either current or other decorator will be notified to raise TypeError
		# as current "_CHAIN_LAYER_HOOK" will be set to "None" as a trigger. This is the handle for
		# situations when "mutable_layer" is not Transmutable or when decorator implementation doesn't
		# have "_mutable_chain_injection".
		self._CHAIN_LAYER_HOOK = None




	def __call__(self, upper_layer_link :Transmutable =None, **kwargs)-> Transmutable :

		if		self._CHAIN_LAYER_HOOK is None: raise TypeError("Not Transmutable transmutation")
		return	self._CHAIN_LAYER_HOOK(upper_layer_link, **kwargs)








class ControlledTransmutation(Transmutable):

	"""
		Pygwarts core class, that serves as a super class for two layers decorators (with arguments).
		The logic behind such decorator the same as for Transmutation, except one thing which allows
		arguments to be included to transmutation only - the decorator initiation shift.
		For ControlledTransmutation, __init__ must handle arguments. By default it takes all positional
		arguments and store it to a "layer_args" field, then in takes all keyword arguments and creates
		fields mapping. Those arguments-fields access might be provided in "_mutable_chain_injection"
		callable, that must handles transmutation of decorated class. Extracting those arguments-fields
		to a "_mutable_chain_injection" namespace variable will grant use of it for transmutated class.
		As this logic is very simple and prone to some misleading about which arguments are positional
		and which are key-word only, and as the actual implementation of any ControlledTransmutation
		decorator must rely on this, it is the best practice to implement dissent __init__ for every
		ControlledTransmutation decorator, to ensure maintaining of neccessary arguments manipulations.
		For ControlledTransmutation, __call__ will be a part of initiation where it accepts the decorated
		entity as "mutable_layer" and considers it's transmutation - whether it is Transmutable to proceed
		with or it is not Transmutable to stop the initiation. Basically ControlledTransmutation __call__
		is a Transmutation __init__, except for ControlledTransmutation it is a termination of initiation
		so it must return transmutated class or make a signal for upper layer that not a Transmutable
		object was encountered, so in this case it returns special function which when invoked will raise
		TypeError with corresponding message. It is complied to a mutable chain initiation, which for every
		hooked Transmutable will try to initiate it, so if for Transmutation there is __call__ that handles
		exactly this situation and might raise TypeError right away, ControlledTransmutation must return
		something in it's __call__, and then that something will be invoked in mutable chain initiation,
		so ControlledTransmutation either returns transmutated Transmutable or a callable which will
		raise TypeError upon it's call.
	"""

	def __init__(self, *args, **kwargs):


		self.layer_args = args
		for k,v in kwargs.items(): setattr(self, k,v)


	def __call__(self, mutable_layer :Transmutable) -> Transmutable | Callable[[Any],TypeError] :

		if	callable(getattr(self, "_mutable_chain_injection", None)):
			match (mutation := getattr(mutable_layer, "_CHAIN_LAYER_HOOK", False)):

				# This case considers decorated "mutable_layer" is another Transmutation that wrapped it's
				# "mutable_layer" Transmutable in it's "_CHAIN_LAYER_HOOK".
				case type() if hasattr(mutation, "_MUTABLE_CHAIN_HOOK"):
					return	self._mutable_chain_injection(mutable_layer._CHAIN_LAYER_HOOK)


				# Following two cases in charge for Decorated "mutable_layer" is a Transmutation that
				# tried to wrap not Transmutable class and did set it's "_CHAIN_LAYER_HOOK" to None,
				# so during mutable chain initiation TypError must be raised, or decorated "mutable_layer"
				# might be any other Transmutable class. In both cases "mutable_layer" must be inspected
				# further, so it is up to it's "_MUTABLE_CHAIN_HOOK".
				case None:	pass
				case _:		pass


			# At this point decorated "mutable_layer" must be either a Transmutable or not. If it is
			# Transmutable, it might be either the final point where decoration was started, or it might
			# also be another ControlledTransmutation decorator, which returns transmutated Transmutable.
			if	isinstance(mutable_layer, type) and getattr(mutable_layer, "_MUTABLE_CHAIN_HOOK", None):
				return	self._mutable_chain_injection(mutable_layer)


		# This final point ensures either current or other decorator issues the raise of TypeError for
		# the upper layer by returning special callable. This is the handle for situations when
		# "mutable_layer" is not Transmutable or when decorator implementation doesn't have
		# "_mutable_chain_injection".
		return	tentaclifors










from pathlib								import Path
from shutil									import copy2
from shutil									import copyfile
from pygwarts.magical.philosophers_stone	import Transmutable
from pygwarts.hagrid.thrivables				import Tree








class LeafProbe(Transmutable):

	"""
		hagrid.planting object, that represents simulation of file copying operation.
		The purpose of this object is to indicate the intention of real file operations, before they will
		really affect the file system. There two things that this object __call__ actually does: logging
		debug message if target parent folder doesn't exist (to be created beforehand) and logs info message
		for source file probe.
		If a source file doesn't exist at the start of operation, e.g. that might happen when processing
		some not small files in multi-bough Flourish, the process will be stopped with corresponding log.
		This object __call__ accepts arguments that implement the "planting interface":
			origin	- Tree object that is currently planting, used especially for emitting logging messages;
			sprout	- string that represents the current source root;
			branch	- Path object that represents current source folder;
			leaf	- Path object that represents current operation source file;
			bough	- Path object that represents current operation destination parent folder.
	"""

	def __call__(self, origin :Tree, sprout :str, branch :Path, leaf :Path, bough :Path):

		if	not leaf.is_symlink():

			target	= bough.joinpath(leaf.name)
			shoot	= target.parent


			if	leaf.is_file():
				if	not shoot.is_dir() : origin.loggy.debug(f"Bough \"{shoot}\" probe")


				origin.loggy.info(f"Leaf \"{target}\" probe")
			else:
				origin.loggy.debug(f"Branch \"{leaf}\" not located")
		else:	origin.loggy.debug(f"Symbolic leaf skipped")








class LeafGrowth(Transmutable):

	"""
		hagrid.planting object, that represents file copying operation.
		The whole process divided to 2 stages: first is the check for target file path folder and
		creating it if it doesn't exist, second is the actual copying.
		The copy function that used (shutil.copyfile) doesn't preserve source file meta data.
		No return value for __call__ is assumed, also no handling for any Exception, that might be raised
		by copy function or by parent directory creation (Path.mkdir) function, that must be handled
		by invoker object, e.g. Bloom object.
		If a source file doesn't exist at the start of operation, e.g. that might happen when processing
		some not small files in multi-bough Flourish, the process will be stopped with corresponding log.
		This object's __call__ accepts arguments that implement the "planting interface":
			origin	- Tree object that is currently planting, used especially for emitting logging messages;
			sprout	- string that represents the current source root;
			branch	- Path object that represents current source folder;
			leaf	- Path object that represents current operation source file;
			bough	- Path object that represents current operation destination parent folder.
	"""

	def __call__(self, origin :Tree, sprout :str, branch :Path, leaf :Path, bough :Path):

		if	not leaf.is_symlink():

			target	= bough.joinpath(leaf.name)
			shoot	= target.parent


			if	leaf.is_file():
				if	not shoot.is_dir() : shoot.mkdir(parents=True)


				copyfile(leaf, target, follow_symlinks=False)
				origin.loggy.info(f"Grown leaf \"{target}\"")
			else:
				origin.loggy.debug(f"Branch \"{leaf}\" not located")
		else:	origin.loggy.debug(f"Symbolic leaf skipped")








class LeafMove(Transmutable):

	"""
		hagrid.planting object, that represents file moving operation.
		The whole process divided to 3 stages: first is the check for target file path folder and
		creating it if it doesn't exist; second is the file copying and third is removing of the source
		file in cause of successful copying.
		It is important to note, that copy function that used (shutil.copyfile) doesn't preserve source
		file meta data, and must raise Exception in case of any failure of copying, cause the condition
		to remove source file considers success of previous operation the existence of target file, not
		matter it's modification time, so source might be removed.
		No return value for __call__ is assumed, also no handling for any Exception, that might be raised
		by copy function or by parent directory creation (Path.mkdir) function, or by source removing
		(Path.unlink) function that must be handled by invoker object, e.g. Bloom object.
		If a source file doesn't exist at the start of operation, e.g. that might happen when processing
		some not small files in multi-bough Flourish, the process will be stopped with corresponding log.
		This object's __call__ accepts arguments that implement the "planting interface":
			origin	- Tree object that is currently planting, used especially for emitting logging messages;
			sprout	- string that represents the current source root;
			branch	- Path object that represents current source folder;
			leaf	- Path object that represents current operation source file;
			bough	- Path object that represents current operation destination parent folder.
		IMPORTANT: the unit tests for this object must be inspected for implementation complines,
		cause some tests directly imitates this object behavior.
	"""

	def __call__(self, origin :Tree, sprout :str, branch :Path, leaf :Path, bough :Path):

		if	not leaf.is_symlink():

			target	= bough.joinpath(leaf.name)
			shoot	= target.parent


			if	leaf.is_file():
				if	not shoot.is_dir() : shoot.mkdir(parents=True)


				copyfile(leaf, target, follow_symlinks=False)


				if	target.is_file():

					leaf.unlink()
					origin.loggy.info(f"Moved leaf \"{target}\"")
				else:
					origin.loggy.info(f"Leaf \"{target}\" move failed")
			else:	origin.loggy.debug(f"Branch \"{leaf}\" not located")
		else:		origin.loggy.debug(f"Symbolic leaf skipped")








class LeafClone(Transmutable):

	"""
		hagrid.planting object, that represents operation of copying a file with it's meta data.
		The whole process divided to 3 stages: first is the check for target file path folder and
		creating it if it doesn't exist, second is the actual copying of file and it's meta data,
		third is optional and occurs when the second stage causes Exception raise.
		The copy function that used (shutil.copy2) will try to preserve source file meta data, but this
		process might fail, for example, when doing it between different file systems. In such cases the
		file will be first copied and then file meta data change will be given a try. If this process
		will lead to Exception, like any other option that will cause Exception, it will be tried to be
		handled in place. The target file existent will be considered and it's modification time will be
		compared with modification time before start of operation (if existed for that moment). If target
		file exist and it's modification time greater than value at the start (which defaulted to 0) the
		planting considered half-successful and corresponding info message will be logged, that it is
		grown instead of moved. If any of the above conditions not met, operation considered failed and
		such Exception will be propagated to the invoker, e.g. Bloom object.
		No return value for __call__ is assumed.
		If a source file doesn't exist at the start of operation, e.g. that might happen when processing
		some not small files in multi-bough Flourish, the process will be stopped with corresponding log.
		This object's __call__ accepts arguments that implement the "planting interface":
			origin	- Tree object that is currently planting, used especially for emitting logging messages;
			sprout	- string that represents the current source root;
			branch	- Path object that represents current source folder;
			leaf	- Path object that represents current operation source file;
			bough	- Path object that represents current operation destination parent folder.
		IMPORTANT: the unit tests for this object must be inspected for implementation complines,
		cause some tests directly imitates this object behavior.
	"""

	def __call__(self, origin :Tree, sprout :str, branch :Path, leaf :Path, bough :Path):

		if	not leaf.is_symlink():

			target	= bough.joinpath(leaf.name)
			shoot	= target.parent
			state	= target.stat().st_mtime if target.is_file() else 0


			if	leaf.is_file():
				if	not shoot.is_dir() : shoot.mkdir(parents=True)


				try:

					copy2(leaf, target, follow_symlinks=False)
					origin.loggy.info(f"Cloned leaf \"{target}\"")


				except:
					if	target.is_file():
						if	state <target.stat().st_mtime:

							origin.loggy.info(f"Leaf \"{target}\" cloning stuck, grown instead")
							return
					raise
			else:	origin.loggy.debug(f"Branch \"{leaf}\" not located")
		else:		origin.loggy.debug(f"Symbolic leaf skipped")








class LeafPush(Transmutable):

	"""
		hagrid.planting object, that represents operation of moving a file with it's meta data.
		The whole process divided to 4 stages: first is the check for target file path folder and
		creating it if it doesn't exist; second is the actual moving of file and it's meta data; third is
		the removing of the source file in cause of successful copying and fourth is optional and occurs
		when the second or third stages causes Exception raise.
		The copy function that used (shutil.copy2) will try to preserve source file meta data, and must
		raise Exception in case of any failure of copying, cause the condition to remove source file
		considers success of previous operation the existence of target file, not matter it's modification
		time, so source might be removed. Eventually the process described might fail, for example, when
		doing it between different file systems. In such cases the file will be first copied and then
		file meta data change will be given a try. If the process of copying file with meta data will lead
		to Exception, like any other option that will cause Exception, it will be tried to be handled
		in place. The target file existent will be considered and it's modification time will be compared
		with modification time before start of operation (if existed for that moment). If target file exist
		and it's modification time greater than value at the start (which defaulted to 0) the planting
		considered half-successful and corresponding info message will be logged, that it is grown instead
		of moved. If any of the above conditions not met, operation considered failed and such Exception
		will be propagated to the invoker, e.g. Bloom object.
		No return value for __call__ is assumed.
		If a source file doesn't exist at the start of operation, e.g. that might happen when processing
		some not small files in multi-bough Flourish, the process will be stopped with corresponding log.
		This object's __call__ accepts arguments that implement the "planting interface":
			origin	- Tree object that is currently planting, used especially for emitting logging messages;
			sprout	- string that represents the current source root;
			branch	- Path object that represents current source folder;
			leaf	- Path object that represents current operation source file;
			bough	- Path object that represents current operation destination parent folder.
		IMPORTANT: the unit test for this object must be inspected for implementation complines,
		cause some tests directly imitates this object behavior.
	"""

	def __call__(self, origin :Tree, sprout :str, branch :Path, leaf :Path, bough :Path):

		if	not leaf.is_symlink():

			target	= bough.joinpath(leaf.name)
			shoot	= target.parent
			state	= target.stat().st_mtime if target.is_file() else 0


			if	leaf.is_file():
				if	not shoot.is_dir() : shoot.mkdir(parents=True)


				try:

					copy2(leaf, target, follow_symlinks=False)


					if	target.is_file() and leaf.stat().st_mtime == target.stat().st_mtime:

						leaf.unlink()
						origin.loggy.info(f"Pushed leaf \"{target}\"")
					else:
						origin.loggy.info(f"Leaf \"{target}\" push failed")


				except:
					if	target.is_file():
						if	state <target.stat().st_mtime:

							origin.loggy.info(f"Leaf \"{target}\" pushing stuck, grown instead")
							return
					raise
			else:	origin.loggy.debug(f"Branch \"{leaf}\" not located")
		else:		origin.loggy.debug(f"Symbolic leaf skipped")










import	os
import	unittest
from	pygwarts.magical.philosophers_stone	import Transmutable
from	pygwarts.tests.irma					import IrmaTestCase
from	pygwarts.irma.contrib				import LibraryContrib
from	pygwarts.irma.access				import LibraryAccess
from	pygwarts.irma.access.bookmarks		import VolumeBookmark
from	pygwarts.irma.access.volume			import LibraryVolume
from	pygwarts.irma.access.annex			import VolumeAnnex
from	pygwarts.irma.access.annex			import LibraryAnnex








class AccessAnnexCases(IrmaTestCase):

	"""
		Testing for fields and edge cases
	"""

	@classmethod
	def tearDownClass(cls):

		if	cls.clean_up:
			if	os.path.isfile(cls.ACCESS_ANNEX): os.remove(cls.ACCESS_ANNEX)

	@classmethod
	def setUpClass(cls): cls.make_loggy_file(cls, cls.ACCESS_ANNEX)
	def test_volume_no_upper_init(self):

		class LonelyVolumeAnnex(VolumeAnnex):
			class loggy(LibraryContrib):

				handler		= self.ACCESS_ANNEX
				init_name	= "volume_no_upper_init"
				init_level	= 10

		with self.assertLogs("volume_no_upper_init", 10) as case_loggy:

			self.test_case = LonelyVolumeAnnex()
			self.test_case([])

		self.assertIn("DEBUG:volume_no_upper_init:Library not found", case_loggy.output)




	def test_volume_annex(self):
		class LonelyVolume(LibraryVolume):

			class B1(VolumeBookmark):	trigger = "trigger1"
			class B2(VolumeBookmark):	trigger = "trigger2"
			class B3(VolumeBookmark):	trigger = "trigger3"
			class B4(VolumeBookmark):	trigger = "trigger4"
			class B5(VolumeBookmark):	trigger = "trigger5"
			class Annex(VolumeAnnex):	pass
			class loggy(LibraryContrib):

				handler		= self.ACCESS_ANNEX
				init_name	= "volume_annex"
				init_level	= 10

		self.test_case = LonelyVolume()
		self.test_case.B1(lambda _ : "OOH EEH\n", "inducer")
		self.test_case.B2(lambda _ : "OOH AH AH\n", "inducer")
		self.test_case.B3(lambda _ : "TING TANG\n", "inducer")
		self.test_case.B4(lambda _ : "WALLA WALLA BING BANG\n", "inducer")
		self.test_case.B5(lambda _ : "", "inducer")

		with self.assertLogs("volume_annex", 10) as case_loggy:
			self.assertEqual(

				self.test_case.Annex(self.test_case.keysof("bookmark")),
				"OOH EEH\nOOH AH AH\nTING TANG\nWALLA WALLA BING BANG\n"
			)

		self.assertIn(f"DEBUG:volume_annex:Fetching volume {self.test_case} annex", case_loggy.output)
		self.assertIn(f"DEBUG:volume_annex:{self.test_case.B1} bookmark view 8 symbols", case_loggy.output)
		self.assertIn(f"DEBUG:volume_annex:{self.test_case.B2} bookmark view 10 symbols", case_loggy.output)
		self.assertIn(f"DEBUG:volume_annex:{self.test_case.B3} bookmark view 10 symbols", case_loggy.output)
		self.assertIn(f"DEBUG:volume_annex:{self.test_case.B4} bookmark view 22 symbols", case_loggy.output)
		self.assertIn(f"DEBUG:volume_annex:Volume {self.test_case} annex 50 symbols", case_loggy.output)




	def test_volume_empty_annex(self):
		class LonelyVolume(LibraryVolume):

			class B1(VolumeBookmark):	trigger = "trigger1"
			class B2(VolumeBookmark):	trigger = "trigger2"
			class B3(VolumeBookmark):	trigger = "trigger3"
			class B4(VolumeBookmark):	trigger = "trigger4"
			class B5(VolumeBookmark):	trigger = "trigger5"
			class Annex(VolumeAnnex):	pass
			class loggy(LibraryContrib):

				handler		= self.ACCESS_ANNEX
				init_name	= "volume_empty_annex"
				init_level	= 10

		inducer1 = lambda _ : 42
		inducer2 = lambda _ : ...
		self.test_case = LonelyVolume()
		self.test_case.B1("OOH EEH", "inducer")
		self.test_case.B2(True, "inducer")
		self.test_case.B3(inducer1, "inducer")
		self.test_case.B4(inducer2, "inducer")
		self.test_case.B5(lambda _ : "", "inducer")

		with self.assertLogs("volume_empty_annex", 10) as case_loggy:
			self.assertIsNone(self.test_case.Annex(self.test_case.keysof("bookmark")))

		self.assertIn(f"DEBUG:volume_empty_annex:Fetching volume {self.test_case} annex", case_loggy.output)
		self.assertIn(

			"DEBUG:volume_empty_annex:OOH EEH failed due to TypeError: 'str' object is not callable",
			case_loggy.output
		)
		self.assertIn(

			"DEBUG:volume_empty_annex:True failed due to TypeError: 'bool' object is not callable",
			case_loggy.output
		)
		self.assertIn(

			f"DEBUG:volume_empty_annex:{inducer1} failed due to TypeError: "
			"can only concatenate str (not \"int\") to str",
			case_loggy.output
		)
		self.assertIn(

			f"DEBUG:volume_empty_annex:{inducer2} failed due to TypeError: "
			"can only concatenate str (not \"ellipsis\") to str",
			case_loggy.output
		)
		self.assertIn(f"DEBUG:volume_empty_annex:No view volume {self.test_case}", case_loggy.output)
		self.assertEqual(
			case_loggy.output.count(f"DEBUG:volume_empty_annex:No view volume {self.test_case}"),5
		)
		self.assertIn(f"DEBUG:volume_empty_annex:Volume {self.test_case} annex 0 symbols", case_loggy.output)




	def test_volume_not_viewable_annex(self):
		class LonelyVolume(LibraryVolume):

			class Annex(VolumeAnnex):	pass
			class B1(VolumeBookmark):	trigger = "trigger1"
			class loggy(LibraryContrib):

				handler		= self.ACCESS_ANNEX
				init_name	= "volume_not_viewable_annex"
				init_level	= 10

		self.test_case = LonelyVolume()

		with self.assertLogs("volume_not_viewable_annex", 10) as case_loggy:
			for unviewable in (

				"view", 42, .69, True, False, None, ...,
				[ "view" ],( "view", ),{ "view" },{ "view": "view" }
			):
				with self.subTest(view=unviewable):

					setattr(self.test_case.B1, "view", unviewable)
					self.assertIsNone(self.test_case.Annex(self.test_case.keysof("bookmark")))

			del self.test_case[self.test_case.B1]
			self.test_case.loggy.warning("PART II")

			for nobookmark in "view", 42, .69, True, False, None, ...,( "view", ):
				self.test_case(nobookmark, "bookmark")
			self.assertIsNone(self.test_case.Annex(self.test_case.keysof("bookmark")))

		self.assertIn(
			f"DEBUG:volume_not_viewable_annex:Fetching volume {self.test_case} annex", case_loggy.output
		)
		self.assertIn(
			f"DEBUG:volume_not_viewable_annex:{self.test_case.B1} is not viewable", case_loggy.output
		)
		self.assertIn(f"DEBUG:volume_not_viewable_annex:Invalid bookmark view", case_loggy.output)
		self.assertIn(f"DEBUG:volume_not_viewable_annex:Invalid bookmark 42", case_loggy.output)
		self.assertIn(f"DEBUG:volume_not_viewable_annex:Invalid bookmark 0.69", case_loggy.output)
		self.assertIn(f"DEBUG:volume_not_viewable_annex:Invalid bookmark True", case_loggy.output)
		self.assertIn(f"DEBUG:volume_not_viewable_annex:Invalid bookmark False", case_loggy.output)
		self.assertIn(f"DEBUG:volume_not_viewable_annex:Invalid bookmark None", case_loggy.output)
		self.assertIn(f"DEBUG:volume_not_viewable_annex:Invalid bookmark Ellipsis", case_loggy.output)
		self.assertIn(f"DEBUG:volume_not_viewable_annex:Invalid bookmark ('view',)", case_loggy.output)
		self.assertIn(
			f"DEBUG:volume_not_viewable_annex:Volume {self.test_case} annex 0 symbols", case_loggy.output
		)
		self.assertEqual(
			case_loggy.output.count(
				f"DEBUG:volume_not_viewable_annex:Fetching volume {self.test_case} annex"
			),	12
		)
		self.assertEqual(
			case_loggy.output.count(
				f"DEBUG:volume_not_viewable_annex:{self.test_case.B1} is not viewable"
			),	11
		)
		self.assertEqual(
			case_loggy.output.count(
				f"DEBUG:volume_not_viewable_annex:Volume {self.test_case} annex 0 symbols"
			),	12
		)




	def test_volume_not_string_views_annex(self):
		class LonelyVolume(LibraryVolume):

			class Annex(VolumeAnnex):	pass
			class B1(VolumeBookmark):

				trigger = "trigger1"
				def view(self, volume :LibraryVolume): return 42

			class B2(VolumeBookmark):

				trigger = "trigger2"
				def view(self, volume :LibraryVolume): return .69

			class B3(VolumeBookmark):

				trigger = "trigger3"
				def view(self, volume :LibraryVolume): return True

			class B4(VolumeBookmark):

				trigger = "trigger4"
				def view(self, volume :LibraryVolume): return False

			class B5(VolumeBookmark):

				trigger = "trigger5"
				def view(self, volume :LibraryVolume): return ...

			class B6(VolumeBookmark):

				trigger = "trigger6"
				def view(self, volume :LibraryVolume): return print

			class B7(VolumeBookmark):

				trigger = "trigger7"
				def view(self, volume :LibraryVolume): return Transmutable

			class B8(VolumeBookmark):

				trigger = "trigger8"
				def view(self, volume :LibraryVolume): return ( "view", )

			class B9(VolumeBookmark):

				trigger = "trigger9"
				def view(self, volume :LibraryVolume): return [ "view" ]

			class B10(VolumeBookmark):

				trigger = "trigger10"
				def view(self, volume :LibraryVolume): return { "view" }

			class B11(VolumeBookmark):

				trigger = "trigger11"
				def view(self, volume :LibraryVolume): return { "view": "view" }

			class B12(VolumeBookmark):

				trigger = "trigger12"
				def view(self, volume :LibraryVolume): return None

			class loggy(LibraryContrib):

				handler		= self.ACCESS_ANNEX
				init_name	= "volume_not_string_views_annex"
				init_level	= 10

		self.test_case = LonelyVolume()

		with self.assertLogs("volume_not_string_views_annex", 10) as case_loggy:
			self.assertIsNone(self.test_case.Annex(self.test_case.keysof("bookmark")))

		self.assertIn(

			f"DEBUG:volume_not_string_views_annex:Fetching volume {self.test_case} annex",
			case_loggy.output
		)
		self.assertIn(

			f"DEBUG:volume_not_string_views_annex:{self.test_case.B1} view is not a string",
			case_loggy.output
		)
		self.assertIn(

			f"DEBUG:volume_not_string_views_annex:{self.test_case.B2} view is not a string",
			case_loggy.output
		)
		self.assertIn(

			f"DEBUG:volume_not_string_views_annex:{self.test_case.B3} view is not a string",
			case_loggy.output
		)
		self.assertIn(

			f"DEBUG:volume_not_string_views_annex:{self.test_case.B4} view is not a string",
			case_loggy.output
		)
		self.assertIn(

			f"DEBUG:volume_not_string_views_annex:{self.test_case.B5} view is not a string",
			case_loggy.output
		)
		self.assertIn(

			f"DEBUG:volume_not_string_views_annex:{self.test_case.B6} view is not a string",
			case_loggy.output
		)
		self.assertIn(

			f"DEBUG:volume_not_string_views_annex:{self.test_case.B7} view is not a string",
			case_loggy.output
		)
		self.assertIn(

			f"DEBUG:volume_not_string_views_annex:{self.test_case.B8} view is not a string",
			case_loggy.output
		)
		self.assertIn(

			f"DEBUG:volume_not_string_views_annex:{self.test_case.B9} view is not a string",
			case_loggy.output
		)
		self.assertIn(

			f"DEBUG:volume_not_string_views_annex:{self.test_case.B10} view is not a string",
			case_loggy.output
		)
		self.assertIn(

			f"DEBUG:volume_not_string_views_annex:{self.test_case.B11} view is not a string",
			case_loggy.output
		)
		self.assertIn(

			f"DEBUG:volume_not_string_views_annex:{self.test_case.B12} view is None",
			case_loggy.output
		)
		self.assertIn(

			f"DEBUG:volume_not_string_views_annex:Volume {self.test_case} annex 0 symbols",
			case_loggy.output
		)
















	def test_library_no_upper_init(self):

		class LonelyLibraryAnnex(LibraryAnnex):
			class loggy(LibraryContrib):

				handler		= self.ACCESS_ANNEX
				init_name	= "library_no_upper_init"
				init_level	= 10

		with self.assertLogs("library_no_upper_init", 10) as case_loggy:

			self.test_case = LonelyLibraryAnnex()
			self.test_case()

		self.assertIn("DEBUG:library_no_upper_init:Library not found", case_loggy.output)




	def test_library_direct_annex(self):
		class LonelyLibrary(LibraryAccess):

			class LibAnnex(LibraryAnnex): pass
			class loggy(LibraryContrib):

				handler		= self.ACCESS_ANNEX
				init_name	= "library_annex"
				init_level	= 10

			class V1(LibraryVolume):

				class B1(VolumeBookmark): trigger = "trigger1"
				class B2(VolumeBookmark): trigger = "trigger2"
				class Annex(VolumeAnnex): pass
				inrange = r".+"

			class V2(LibraryVolume):

				class B1(VolumeBookmark): trigger = "trigger3"
				class B2(VolumeBookmark): trigger = "trigger4"
				class Annex(VolumeAnnex): pass
				inrange = r".+"

			class V3(LibraryVolume):

				class B1(VolumeBookmark): trigger = "trigger6"
				class B2(VolumeBookmark): trigger = "trigger7"
				class Annex(VolumeAnnex): pass
				inrange = r".+"

			class V4(LibraryVolume):

				class B1(VolumeBookmark): trigger = "trigger8"
				class B2(VolumeBookmark): trigger = "trigger9"
				class Annex(VolumeAnnex): pass
				inrange = r".+"

			class V5(LibraryVolume):

				class B1(VolumeBookmark): trigger = "trigger10"
				class B2(VolumeBookmark): trigger = "trigger11"
				class Annex(VolumeAnnex): pass
				inrange = r".+"


		self.test_case = LonelyLibrary()
		self.test_case.V1.B1(lambda _ : "OOH ", "inducer")
		self.test_case.V1.B2(lambda _ : "EEH\n", "inducer")
		self.test_case.V2.B1(lambda _ : "OOH ", "inducer")
		self.test_case.V2.B1(lambda _ : "AH ", "inducer")
		self.test_case.V2.B2(lambda _ : "AH\n", "inducer")
		self.test_case.V3.B1(lambda _ : "TING ", "inducer")
		self.test_case.V3.B2(lambda _ : "TANG\n", "inducer")
		self.test_case.V4.B1(lambda _ : "WALLA ", "inducer")
		self.test_case.V4.B2(lambda _ : "WALLA ", "inducer")
		self.test_case.V5.B1(lambda _ : "BING ", "inducer")
		self.test_case.V5.B2(lambda _ : "BANG", "inducer")

		# Imitating volume polution by handlers
		self.test_case.V1("some handler gets", 42)
		self.test_case.V2("some handler gets", 42)
		self.test_case.V3("some handler gets", 42)
		self.test_case.V4("some handler gets", 42)
		self.test_case.V5("some handler gets", 42)


		self.assertEqual(self.test_case.LibAnnex(), "OOH EEH\nOOH AH AH\nTING TANG\nWALLA WALLA BING BANG")








	def test_library_direct_annex_no_range(self):
		class LonelyLibrary(LibraryAccess):

			class LibAnnex(LibraryAnnex): pass
			class loggy(LibraryContrib):

				handler		= self.ACCESS_ANNEX
				init_name	= "library_annex_no_range"
				init_level	= 10

			class V1(LibraryVolume):

				class B1(VolumeBookmark): trigger = "trigger1"
				class B2(VolumeBookmark): trigger = "trigger2"
				class Annex(VolumeAnnex): pass

			class V2(LibraryVolume):

				class B1(VolumeBookmark): trigger = "trigger3"
				class B2(VolumeBookmark): trigger = "trigger4"
				class Annex(VolumeAnnex): pass

			class V3(LibraryVolume):

				class B1(VolumeBookmark): trigger = "trigger6"
				class B2(VolumeBookmark): trigger = "trigger7"
				class Annex(VolumeAnnex): pass

			class V4(LibraryVolume):

				class B1(VolumeBookmark): trigger = "trigger8"
				class B2(VolumeBookmark): trigger = "trigger9"
				class Annex(VolumeAnnex): pass

			class V5(LibraryVolume):

				class B1(VolumeBookmark): trigger = "trigger10"
				class B2(VolumeBookmark): trigger = "trigger11"
				class Annex(VolumeAnnex): pass


		self.test_case = LonelyLibrary()
		self.test_case.V1.B1(lambda _ : "OOH ", "inducer")
		self.test_case.V1.B2(lambda _ : "EEH\n", "inducer")
		self.test_case.V2.B1(lambda _ : "OOH ", "inducer")
		self.test_case.V2.B1(lambda _ : "AH ", "inducer")
		self.test_case.V2.B2(lambda _ : "AH\n", "inducer")
		self.test_case.V3.B1(lambda _ : "TING ", "inducer")
		self.test_case.V3.B2(lambda _ : "TANG\n", "inducer")
		self.test_case.V4.B1(lambda _ : "WALLA ", "inducer")
		self.test_case.V4.B2(lambda _ : "WALLA ", "inducer")
		self.test_case.V5.B1(lambda _ : "BING ", "inducer")
		self.test_case.V5.B2(lambda _ : "BANG", "inducer")


		with self.assertLogs("library_annex_no_range", 10) as case_loggy:
			self.assertIsNone(self.test_case.LibAnnex())

		self.assertIn(

			f"DEBUG:library_annex_no_range:Fetching library {self.test_case} annex",
			case_loggy.output
		)
		self.assertIn(

			f"INFO:library_annex_no_range:Valid {self.test_case.V1} range not provided",
			case_loggy.output
		)
		self.assertIn(

			f"INFO:library_annex_no_range:Valid {self.test_case.V2} range not provided",
			case_loggy.output
		)
		self.assertIn(

			f"INFO:library_annex_no_range:Valid {self.test_case.V3} range not provided",
			case_loggy.output
		)
		self.assertIn(

			f"INFO:library_annex_no_range:Valid {self.test_case.V4} range not provided",
			case_loggy.output
		)
		self.assertIn(

			f"INFO:library_annex_no_range:Valid {self.test_case.V5} range not provided",
			case_loggy.output
		)
		self.assertIn(

			f"DEBUG:library_annex_no_range:Library {self.test_case} annex 0 symbols",
			case_loggy.output
		)








if	__name__ == "__main__" : unittest.main(verbosity=2)










from typing						import List
from time						import sleep
from threading					import Thread
from multiprocessing			import current_process
from pygwarts.magical.spells	import flagrate
from pygwarts.irma.contrib		import LibraryContrib








class ContribInterceptor:

	"""
		LibraryContrib decorators super class. Implements the mutable chain injection for a LibraryContrib
		children with "CONTRIB_HOOK" field and it's corresponding processing in decoration time. As decorated
		layer in __init__ must be either LibraryContrib or it's another decorator, "CONTRIB_HOOK" will allow
		it's recognition and further handling.
	"""

	CONTRIB_HOOK = True

	def __init__(self, layer :LibraryContrib) -> LibraryContrib :

		match layer:

			case	type() if getattr(layer, "CONTRIB_HOOK", None):		self.layer = layer
			case	object() if getattr(layer, "CONTRIB_HOOK", None):	self.layer = layer()
			case _:	raise TypeError(f"Intercepting invalid object \"{layer}\"")








class STDOUTL(ContribInterceptor):

	"""
		LibraryContrib decorator, that overwrites such LibraryContrib level methods like:
			"info",
			"warning",
			"error",
			"critical",
		to intercept messages and put it to the standard output (print it) before return
		to the LibraryContrib corresponding level method.
	"""

	def __call__(self):
		class Interceptor(self.layer):

			def info(self, message :str):

				print(f"{self.handover_name} INFO : {message}")
				return super().info(message)

			def warning(self, message :str):

				print(f"{self.handover_name} WARNING : {message}")
				return super().warning(message)

			def error(self, message :str):

				print(f"{self.handover_name} ERROR : {message}")
				return super().error(message)

			def critical(self, message :str):

				print(f"{self.handover_name} CRITICAL : {message}")
				return super().critical(message)


		return	Interceptor








class PoolHoist(ContribInterceptor):

	"""
		LibraryContrib main decorator, that implements the hoisting of any strings, which are considered
		to got through the decorated LibraryContrib object, to the pool of threads. Basically strings
		are the logging messages, hoisting is actual intercepting of them on order to put in special
		buffer and pool of threads is the schedule of such buffer termination methods calls. By default
		"PoolHoist" doesn't implement the actual hoist, only the buffer and pool communication, so
		it is possible to construct any level hoisting within other objects, either decorators or even
		in mutable chain. The process of pool interaction releis on following methods:
			buffer_insert	- the method to be invoked for actual intercepting; will accept the message
							string and will maintain the buffer and pool insertion; in total, this method
							must be called for every string that is intended to be hoisted;
			buffer_release	- the method that is a buffer terminator; it is always invoked in a thread
							to implement delayed buffer comparison; as this method returns terminated
							buffer, the best idea is to overwrite it in any descent decorator, or even
							in direct inheritance.
		For every message to be inserted in buffer a unique id number is created to escort it. For the
		very first message hoist, when the buffer is first time extended, the process of buffering is
		turned on and special field's "POOL_STATE" first bit set to 1 will indicate it. Along with it
		the thread with buffer terminator is started, which will delay the actual termination by
		"POOL_TIMER" seconds every time the buffer will be altered. This delay is limited by "POOL_LIMIT".
		All messages hoisting during "buffering" will be the buffer altering that will cause delay. After
		some time that not exceeds "POOL_LIMIT" seconds and buffer changing stopped, or after "POOL_LIMIT"
		exceed, the buffer will terminated.
	"""


	def __call__(self):
		class Interceptor(self.layer):


			def __init__(self, *args, **kwargs):
				super().__init__(*args, **kwargs)


				self.POOL_TIMER	= getattr(super(), "pool_timer", 5)
				self.POOL_LIMIT	= getattr(super(), "pool_limit", 50)
				self.POOL_NAME	= current_process().name
				self.POOL_STATE	= int()
				self.POOL_SET	= set()
				self.BUFFER		= list()




			@property
			def BL(self): return len(self.BUFFER)
			def buffer_insert(self, message :str):

				"""
					Main method, that takes any string, that considered as some message, for example
					emitted at any log level higher than debug, and puts it in the pool with "pool_insert"
					method. For every "buffer_insert" call the unique number "POOL_ID" is created to serve
					as unique identifier for current "message". This "POOL_ID" will be first used to locate
					current pool with "pool_locate" method, which might wipe the whole buffer. After pool
					location the "BUFFER" list will be extended with current "message", along with insertion
					into the pool with "pool_insert" method. "POOL_SET" will be also maintained and all
					messages will be first check if it is already in it, so duplicates will be skipped.
				"""

				if	isinstance(message, str) and (mlen := len(message)) and message not in self.POOL_SET:
					P = self.pool_get_new_id()


					self.pool_locate(P)
					self.POOL_SET.add(message)
					self.BUFFER.append(message)


					self.pool_debug(f"(pool-id-{P}) {mlen} symbol{flagrate(mlen)} message inserted")
					self.pool_debug(f"(pool-id-{P}) buffer extended to {self.BL} item{flagrate(self.BL)}")
					self.pool_insert(P)




			def buffer_release(self, P :int) -> List[str] :

				"""
					Main method that acts as a buffer terminator. When invoked it is assumed that "BUFFER"
					was altered to have the first message, so the cycle of buffer comparison is started.
					This cycle checks the buffer size and buffer time, which both are set to "BUFFER" items
					count and "POOL_TIMER" seconds corresponding. On every iteration of cycle, after the
					"POOL_TIMER" timeout the buffer size again is set to "BUFFER" items count and buffer
					time incremented by "POOL_TIMER". The cycle goes on this way until either buffer size
					will have no changes between iterations, or the buffer time exceeds "POOL_LIMIT" seconds.
					When the cycle ends, "POOL_STATE" first bit will be set to 0 to inform that buffering
					is over. The "POOL_SET" and "BUFFER" will be wiped and it's content will be returned.
				"""

				buffer_size = 0
				buffer_time = 0


				while buffer_size <self.BL and buffer_time <self.POOL_LIMIT:


					buffer_size = self.BL
					buffer_time += self.POOL_TIMER


					sleep(self.POOL_TIMER)
					if self.POOL_TIMER <buffer_time : self.pool_debug(f"(pool-id-{P}) timer extended")


				self.pool_debug(f"(pool-id-{P}) was {buffer_time} second{flagrate(buffer_time)}")
				self.pool_debug(f"(pool-id-{P}) was {buffer_size} item{flagrate(buffer_size)}")


				self.POOL_STATE |= 1
				self.POOL_STATE ^= 1


				buffer_dump	= self.BUFFER


				self.POOL_SET = set()
				self.BUFFER	= list()


				return	buffer_dump








			def pool_debug(self, message :str):

				"""
					PoolHoist inner debug message emitter. Implements the Transmutable style of logging,
					when it is "handovered" the "loggy" object, which in this situation is itself, and
					then emits message. The handovered name in this case is not taken from any layer object
					as in mutable chain, but is strictly defined as "loggy.pool". As any overwriting of
					"debug" method functionality will affect current inner logging method, and as debug
					flow might be extremely vast, the best idea is to never intercept it.
				"""

				self.handover("loggy.pool", assign=False)
				self.debug(message)




			def pool_get_new_id(self) -> int :

				"""
					Helper method that maintains the generation of unique id numbers to be used for
					pool operations. Starts with 0 and increments every time called.
				"""

				current_id = getattr(self, "POOL_ID_GENERATOR", 0)
				self.POOL_ID_GENERATOR = current_id +1

				return current_id




			def pool_insert(self, P :int):

				"""
					The crucial method of pool, which serves as a thread dispatcher. Accepts the unique
					pool id number "P" and first of all checks if there is any already started thread by
					"POOL_STATE" 1 bit set. If 1 bit set it means the "buffering" state and all messages
					must be handled by already started thread. Otherwise the "P" id will be used with
					new thread start. The new thread starts the "buffer_release" method which will gather
					messages while "buffering" is on.
				"""

				if	self.POOL_STATE &1 == 0:
					self.POOL_STATE ^= 1

					self.pool_debug(f"(pool-id-{P}) starting new thread")
					Thread(target=self.buffer_release, kwargs={ "P": P }).start()




			def pool_locate(self, P :int):

				"""
					Core method that allows to detect pool migration between processes. If the process
					where PoolHoist was originally started will be forked, which will be detected by process
					name changes, when the parent process had thread started, it means the child process will
					inherit the "POOL_STATE" set to "buffering", which means there is already working thread.
					But when such thread, which is "buffer_release" method, will be stopped, it will happen
					only in parent process and the child "POOL_STATE" will stay the same. That will restrict
					thread starting in child process. This method solves this by locating current process
					and managing "POOL_STATE" when forked. Also the inherited buffer and set will be wiped.
					Accepts the unique pool id number "P" to be used for maintaining.
				"""

				if	(current_name := current_process().name) != self.POOL_NAME :

					self.pool_debug(f"(pool-id-{P}) switching from {self.POOL_NAME} to {current_name}")
					self.pool_debug(f"(pool-id-{P}) wiping {self.BL} item{flagrate(self.BL)} buffer")

					self.POOL_STATE |= 1
					self.POOL_STATE ^= 1

					self.POOL_NAME	= current_name
					self.BUFFER		= list()
					self.POOL_SET	= set()


		return	Interceptor










import	unittest
from	pygwarts.tests.hagrid.unit_hagrid_init		import *
from	pygwarts.tests.hagrid.unit_hagrid_order		import *
from	pygwarts.tests.hagrid.unit_hagrid_sifting	import *
from	pygwarts.tests.hagrid.unit_hagrid_peels		import *
from	pygwarts.tests.hagrid.unit_hagrid_BlindPeek	import *
from	pygwarts.tests.hagrid.unit_hagrid_DraftPeek	import *
from	pygwarts.tests.hagrid.unit_hagrid_flourish	import *
from	pygwarts.tests.hagrid.unit_hagrid_easy1		import *
from	pygwarts.tests.hagrid.unit_hagrid_easy2		import *
from	pygwarts.tests.hagrid.unit_hagrid_easy3		import *
from	pygwarts.tests.hagrid.unit_hagrid_easy4		import *
from	pygwarts.tests.hagrid.unit_hagrid_easy5		import *
from	pygwarts.tests.hagrid.unit_hagrid_easy6		import *
from	pygwarts.tests.hagrid.unit_hagrid_easy7		import *
from	pygwarts.tests.hagrid.unit_hagrid_easy8		import *
from	pygwarts.tests.hagrid.unit_hagrid_easy9		import *
from	pygwarts.tests.hagrid.unit_hagrid_easy10	import *
from	pygwarts.tests.hagrid.unit_hagrid_easy11	import *
from	pygwarts.tests.hagrid.unit_hagrid_easy12	import *
from	pygwarts.tests.hagrid.unit_hagrid_easy13	import *
from	pygwarts.tests.hagrid.unit_hagrid_easy14	import *
from	pygwarts.tests.hagrid.unit_hagrid_easy15	import *
from	pygwarts.tests.hagrid.unit_hagrid_medium1	import *
from	pygwarts.tests.hagrid.unit_hagrid_medium2	import *
from	pygwarts.tests.hagrid.unit_hagrid_medium3	import *
from	pygwarts.tests.hagrid.unit_hagrid_medium4	import *
from	pygwarts.tests.hagrid.unit_hagrid_medium5	import *
from	pygwarts.tests.hagrid.unit_hagrid_medium6	import *
from	pygwarts.tests.hagrid.unit_hagrid_medium7	import *
from	pygwarts.tests.hagrid.unit_hagrid_medium8	import *
from	pygwarts.tests.hagrid.unit_hagrid_medium9	import *
from	pygwarts.tests.hagrid.unit_hagrid_medium10	import *
from	pygwarts.tests.hagrid.unit_hagrid_medium11	import *
from	pygwarts.tests.hagrid.unit_hagrid_medium12	import *








if	__name__ == "__main__" : unittest.main(verbosity=2)







